"""Definition of the functionality of a Backend within the framework

A Backend is designed to be the single central place for persistent storage within the framework. It is responsible
for managing all changes within the state of a project. The backend touches almost all other primitives within
the framework. To gain a better understanding of its larger role with the framework, refer to the documentation
about the design of the framework (link).

This file provides the definition of the functionality a backend should implement, but it does not provide 
a concrete implementation. 

This file also contains the definition of how to configure and dynamically load a backend class. To denote a 
concrete implementation, a python `module` and `class` must be provide. These should both be strings. They are
used to load a class that inherits from the base `Backend` class. It also takes in a generic dictionary to be 
passed as configuration into the dynamicall loaded class.

"""
import inspect
from typing import Dict, Tuple, Any, List
from pydantic import BaseModel

from .components import Component_Difference, ComponentModel

from .resource import Resource_Reference_Difference, ResourceModel, Resource_Difference
from .resource_state import Resource_State

from ..utils.module_loader import import_module


class Backend_Configuration(BaseModel):
    python_module: str
    python_class: str
    config: Dict

    def __init__(
        __pydantic_self__, python_module: str, python_class: str, config: Dict
    ) -> None:
        """
        Represents the data needed to create a new cdev backend:

        Parameters:
            python_module: The name of the python module to load as the backend
            config: configuration option for the backend

        """

        super().__init__(
            **{
                "python_module": python_module,
                "python_class": python_class,
                "config": config,
            }
        )


class Backend:
    def __init__(self, **kwargs) -> None:
        raise NotImplementedError

    # Api for working with Resource States
    def create_resource_state(
        self, name: str, parent_resource_state_uuid: str = None
    ) -> str:
        """
        Create a new resource state within this stored state.

        Args:
            parent_resource_state_uuid (str): The uuid of the parent resource state.
            name (str): the name of this resource state.

        Returns:
            uuid (str): The uuid of the newly created resource state.

        Raises:
            ResourceStateAlreadyExists

        """
        raise NotImplementedError

    def delete_resource_state(self, resource_state_uuid: str):
        """
        Delete a resource state within this store state.

        Args:
            resource_state_uuid (str): The uuid of the resource state to delete

        Raises:
            ResourceStateDoesNotExist
            ResourceStateNotEmpty
        """
        raise NotImplementedError

    def get_resource_state(self, resource_state_uuid: str) -> Resource_State:
        """
        Load a resource state from the stored state.

        Args:
            resource_state_uuid (str): The uuid of the desired resource state

        Raises:
            ResourceStateDoesNotExist
        """
        raise NotImplementedError

    def get_top_level_resource_states(self) -> List[Resource_State]:
        """
        List all the top level resource states for this stored state

        Returns:
            resource_states (List[Resource_States]): Top level resource states.

        Raises:
            ResourceStateDoesNotExist

        """
        raise NotImplementedError


    def get_component(
        self, resource_state_uuid: str, component_name: str
    ) -> ComponentModel:
        """
        Get a component from the resource state.

        Args:
            resource_state_uuid (str): The resource state that this component will be in.
            component_name (str): Name of the component

        Returns:
            component (ComponentModel): The data for the requested component

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
        """
        raise NotImplementedError


    def update_component(
        self, resource_state_uuid: str, component_difference: Component_Difference
    ):
        """Make an update to the component. 

        Args:
            resource_state_uuid (str): The resource state that the change is in.
            component_difference (Component_Difference): The difference in the component to be made.

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
        """
        pass


    def get_component_uuid(self, resource_state_uuid: str, component_name: str) -> str:
        """
        Get the unique namespace for this resource state and component. 

        Args:
            resource_state_uuid (str): The resource state that this component will be in.
            component_name (str): Name of the component

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist

        Returns:
            identifier (str): A unique identifier for this namespace
        """

    # Api for changing individual Resources
    # The resource state needs to know when a mapper will be attempting to update a cdev resource. It is in charge of
    # determing if the current resource state is capable of handling a change in the resource.
    def create_resource_change_transaction(
        self, resource_state_uuid: str, component_name: str, diff: Resource_Difference
    ) -> Tuple[str, str]:
        """
        Create in the Resource State the desire to change a particular resource. If the resource state can currently handle creating
        this change, it will return a base idempotency token that can be used to construct idempotency tokens for deploying the underlying
        cloud resources. If the resource state can not handle the change, it will throw an error.

        Args:
            resource_state_uuid (str): The resource state for this resource change.
            component_name (str): The component this resource change is occuring in.
            diff (Resource_Difference): The desired change in the resource.

        Returns:
            transaction_token (str): The transaction token to be used by the mapper when deploying the resource. This token can be used to give to a cloud
            provider as a idempotency token.
            namespace_token (str): A suffix that can be added to deployed cloud resources that acts as a namespace for the resource within the cloud. 

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
        """
        raise NotImplementedError

    def complete_resource_change(
        self,
        resource_state_uuid: str,
        component_name: str,
        diff: Resource_Difference,
        transaction_token: str,
        cloud_output: Dict,
        resolved_cloud_information: Dict={}
    ):
        """
        Notify the resource state that all changes to a resource have completed successfully. This will cause the resource to
        update the state of the resource to the new state.

        Args:
            resource_state_uuid (str): The resource state for this resource change.
            component_name (str): The component this resource change is occuring in.
            diff (Resource_Difference): The desired change in the resource
            transaction_token (str): Identifying token representing what transaction is being completed
            cloud_output (Dict): Output information from the cloud provider
            resolved_cloud_information (Dict): Information that was resolved from the cloud output of the component to deploy this change. 


        Raises:
            ResourceChangeTransactionDoesNotExist
            ComponentDoesNotExist

        """
        raise NotImplementedError

    def fail_resource_change(
        self,
        resource_state_uuid: str,
        component_name: str,
        diff: Resource_Difference,
        transaction_token: str,
        failed_state: Dict,
    ):
        """
        Notify the resource state that an attempted change to a resource has failed. The provided failed state should encapsulate any needed information
        for a future mapper to recover the state of the resource back into a proper state.

        Args:
            resource_state_uuid (str): The resource state for this resource change.
            component_name (str): The component this resource change is occuring in.
            diff (Resource_Difference): The desired change in the resource
            transaction_token (str): Identifying token representing what transaction is being completed
            failed_state (Dict): A dictionary containing information a mapper could use to resolve the failed state

        Raises:
            ResourceStateDoesNotExist
            ResourceChangeTransactionDoesNotExist
        """
        raise NotImplementedError

    # Api for getting references from other components
    # Components can have references to resources that are managed in other components and it is the responsibility of the backend to resolve those references
    # The backend should be in charge of handling IAM to determine if resolving the reference is possible
    def resolve_reference_change(
        self,
        resource_state_uuid: str,
        diff: Resource_Reference_Difference,
    ):
        """
        Either reference or dereference a resource from a different component.

        Args:
            resource_state_uuid (str): The resource state for this resource change.
            component_name (str): The component this resource change is occuring in.
            diff (Resource_Difference): The desired change in the resource

        Raises:
            ResourceReferenceError
        """
        raise NotImplementedError

    # Api for working with a resource states failed resource updates
    # We can either update the failed state by the mapper attempting to fix the underlying issue, or
    # We can recover the resource to either the new state or previous state by the mapper fixing the issues, or
    # We can just delete the failure from our failed state (not recommended unless you know what you are doing because it can leave resources in the cloud)
    def change_failed_state_of_resource_change(
        self, resource_state_uuid: str, transaction_token: str, new_failed_state: Dict
    ):
        """
        Update the failed state of a 'resource change'.

        Args:
            resource_state_uuid (str): The resource state that this transaction is in.
            transaction_token (str): Identifying token for the failed transaction.
            new_failed_state (Dict): The new failed state of the transaction.

        Raises:
            ResourceStateDoesNotExist
            ResourceChangeTransactionDoesNotExist

        """
        raise NotImplementedError

    def recover_failed_resource_change(
        self,
        resource_state_uuid: str,
        transaction_token: str,
        to_previous_state: bool = True,
    ):
        """
        Recover the state of the change back to the previous state or forward to the new state. Note this will result in the failed resource change being removed from the stored state.

        Args:
            resource_state_uuid (str): The resource state that this transaction is in.
            transaction_token (str): Identifying token for the failed transaction.
            to_previous_state (bool): Bool to decide if the resource should be transitioned to the previous state or new state.

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            ResourceChangeTransactionDoesNotExist
        """
        raise NotImplementedError

    def remove_failed_resource_change(
        self, resource_state_uuid: str, transaction_token: str
    ):
        """
        Completely remove the failed resource change from the stored state. Note that this should be used with caution as it can lead to hanging cloud resources.

        Args:
            resource_state_uuid (str): The resource state that this transaction is in.
            transaction_token (str): Identifying token for the failed transaction.

        Raises:
            ResourceStateDoesNotExist
            ResourceChangeTransactionDoesNotExist
        """
        raise NotImplementedError

    # Api for getting information about a resource from the backend
    def get_resource_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
    ) -> ResourceModel:
        """
        Get the state of a resource from a component based on the name of the resource

        Args:
            resource_state_uuid (str): The resource state for this resource.
            component_name (str): The component this resource is in.
            resource_type: The RUUID of the resource desired
            resource_name: The name of the resource desired

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            ResourceDoesNotExist
        """
        raise NotImplementedError

    def get_resource_by_hash(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_hash: str,
    ) -> ResourceModel:
        """
        Get the state of a resource from a component based on the hash of the resource

        Args:
            resource_state_uuid (str): The resource state for this resource.
            component_name (str): The component this resource is in.
            resource_type (str): The RUUID of the resource desired.
            resource_hash (str): The hash of the resource desired.

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            ResourceDoesNotExist
        """
        raise NotImplementedError

    def get_cloud_output_value_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
        key: str,
    ) -> Any:
        """
        Get an output value from the cloud provider for a resource by the name of the resource. This function also takes an optional function as a parameter
        that will be executed on the output and should return a transformed version of the information.

        Args:
            resource_state_uuid (str): The resource state for this resource.
            component_name (str): The component this resource is in.
            resource_type (str): The RUUID of the resource desired
            resource_name (str): The hash of the resource desired
            key (str): The key for the desired value

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            CloudOutputDoesNotExist
            KeyNotInCloudOutput

        """
        raise NotImplementedError


    def get_cloud_output_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
    ) -> Dict:
        """
        Get full output from the cloud provider for a resource by the name of the resource.

        Args:
            resource_state_uuid (str): The resource state for this resource.
            component_name (str): The component this resource is in.
            resource_type (str): The RUUID of the resource desired
            resource_name (str): The hash of the resource desired
            key (str): The key for the desired value

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            CloudOutputDoesNotExist
            KeyNotInCloudOutput

        """
        raise NotImplementedError

    def get_cloud_output_value_by_hash(
        self,
        resource_state_uuid: str,
        component_uuid: str,
        resource_type: str,
        resource_hash: str,
        key: str,
    ) -> Any:
        """
        Get an output value from the cloud provider for a resource by the name of the resource. This function also takes an optional function as a parameter
        that will be executed on the output and should return a transformed version of the information.

        Args:
            resource_state_uuid (str): The resource state for this resource.
            component_name (str): The component this resource is in.
            resource_type (str): The RUUID of the resource desired.
            resource_hash (str): The hash of the resource desired.
            key (str): The key for the desired value.

        Raises:
            ResourceStateDoesNotExist
            ComponentDoesNotExist
            KeyNotInCloudOutput
            CloudOutputDoesNotExist

        """
        raise NotImplementedError

    # Api for creating differences between workspace and the backend
    # The backend must be responsible for determining the differences to get to a new potential state because the backend
    # is the one that has to implement the changes.
    # In the future, this will also allow for the backend to manage IAM permissions for creating, updating, referencing, etc
    # the actual cdev resources.
    def create_differences(
        self,
        resource_state_uuid: str,
        new_components: List[ComponentModel],
        old_components: List[str],
    ) -> Tuple[
        Component_Difference,  Resource_Difference, Resource_Reference_Difference,
    ]:
        """
        Create the set of differences from a proposed set of components to a provided set of current components identified by their name. This allows the flexibility for working on a particular
        set of components within a resource state.
        """
        raise NotImplementedError


def load_backend(config: Backend_Configuration) -> Backend:
    """Dynamically load a backend

    Load a backend by initializing the provided module and looking for the class within the loaded module.

    Args:
        config (Backend_Configuration): configuration of what backend to load.

    Raises:
        Exception: [description]

    Returns:
        Backend: [description]
    """
    try:
        backend_module = import_module(config.python_module)
    except Exception as e:
        print("Error loading backend module")
        print(f"Error > {e}")

        raise e

    backend_class = None
    for item in dir(backend_module):
        potential_obj = getattr(backend_module, item)
        if (
            inspect.isclass(potential_obj)
            and issubclass(potential_obj, Backend)
            and item == config.python_class
        ):
            backend_class = potential_obj
            break

    if not backend_class:
        print(f"Could not find {config.python_class} in {config.python_module}")
        raise Exception

    try:
        # initialize the backend obj with the provided configuration values
        initialized_obj = potential_obj(**config.config)
    except Exception as e:
        print(f"Could not initialize {potential_obj} Class from config {config.config}")
        raise e

    return initialized_obj



from argparse import ArgumentParser
from typing import Dict, List

import boto3


from core.constructs.commands import BaseCommand, OutputWrapper
from core.default.resources.simple.table import Table
from core.utils.paths import get_full_path_from_workspace_base


from . import utils

RUUID = "cdev::simple::table"


class clear_table(BaseCommand):

    help = """
Clear the current data from a given Table. This should only be used on development tables.   
"""

    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "resource_name", type=str, help="The resource you want to sync data to"
        )

    def command(self, *args, **kwargs):
        """
        Clear all items for the table
        """
        # https://stackoverflow.com/questions/55169952/delete-all-items-dynamodb-using-python

        full_resource_name = kwargs.get("resource_name")
        component_name = full_resource_name.split('.')[0]
        table_resource_name = full_resource_name.split('.')[1]

        cloud_output = utils.get_cloud_output_from_cdev_name(component_name, table_resource_name)
        table_cloud_name = cloud_output.get('table_name')

        dynamo = boto3.resource("dynamodb")
        table = dynamo.Table(table_cloud_name)

        # get the table keys
        tableKeyNames = [key.get("AttributeName") for key in table.key_schema]

        # Only retrieve the keys for each item in the table (minimize data transfer)
        projectionExpression = ", ".join("#" + key for key in tableKeyNames)
        expressionAttrNames = {"#" + key: key for key in tableKeyNames}

        counter = 0
        page = table.scan(
            ProjectionExpression=projectionExpression,
            ExpressionAttributeNames=expressionAttrNames,
        )
        with table.batch_writer() as batch:
            while page["Count"] > 0:
                counter += page["Count"]
                # Delete items in batches
                for itemKeys in page["Items"]:
                    self.stderr.write(f"Removing item {itemKeys}")
                    batch.delete_item(Key=itemKeys)
                # Fetch the next page
                if "LastEvaluatedKey" in page:
                    page = table.scan(
                        ProjectionExpression=projectionExpression,
                        ExpressionAttributeNames=expressionAttrNames,
                        ExclusiveStartKey=page["LastEvaluatedKey"],
                    )
                else:
                    break

        self.stderr.write(f"Deleted {counter}")



from core.constructs.workspace import Workspace


RUUID = "cdev::simple::function"

def get_cloud_id_from_cdev_name(component_name: str, cdev_function_name: str) -> str:
    try:
        ws = Workspace.instance()


        cloud_id = ws.get_backend().get_cloud_output_value_by_name(
            ws.get_resource_state_uuid(),
            component_name,
            RUUID, 
            cdev_function_name, 
            "cloud_id"
        )

        return cloud_id
    except Exception as e:
        print(f"Could not find cloud id")
        print(e)
        return None



"""Set of constructs for creating triggerable events

"""

from typing import Optional, Union

from core.constructs.models import frozendict, ImmutableModel
from core.default.resources.simple.iam import permission_arn_model, permission_model

################
##### Events
################
class event_model(ImmutableModel):
    originating_resource_name: str

    originating_resource_type: str

    hash: str

    granting_permission: Optional[Union[permission_model, permission_arn_model]]



class Event():
    
    def hash(self) -> str:
        raise NotImplementedError

    def render(self,) -> event_model:
        raise NotImplementedError



# Generated as part of Resource Test project template
from cdev.resources.simple.xlambda import simple_function_annotation
from cdev.resources.simple.api import Api
from cdev.resources.simple.object_store import Bucket
from cdev.resources.simple.queue import Queue
from cdev.resources.simple.relational_db import RelationalDB, db_engine
from cdev.resources.simple.static_site import StaticSite
from cdev.resources.simple.table import Table, AttributeDefinition, KeyDefinition, attribute_type, key_type
from cdev.resources.simple.topic import Topic


# Api
myApi = Api("demoapi")
route1 = myApi.route('/hello', 'GET')

# Bucket
bucket = Bucket('demobucket')

# Queue
myQueue = Queue('bigqueue')


# Table
attributes = [AttributeDefinition('id', attribute_type.S)]
keys = [KeyDefinition('id', key_type.HASH)]

dinner_table = Table('dinner', attributes, keys)

# Topic
good_dinner_conversation_topic = Topic('of_discussion', nonce='1')


# Function
@simple_function_annotation("hello_world_function", permissions=[bucket.available_permissions.READ_AND_WRITE_BUCKET])
def hello_world(event, context):
    print('Hellow World!')


"""Utilities to help work with filesystem paths within the framework.


To make projects work on multiple developers machines, it is important that paths that
need to be stored in the state be written as relative paths. The relative paths can be 
started from two places: Workspace and Intermediate.

The Workspace base is the base folder for all information that is unique to the current work 
space. This will be default be the directory that a cli command is issued from. Changing
this value should be done with caution and understanding of how to framework/cli bootstraps
itself.


The Intermediate base is the base folder for any artifacts that are created by the framework
that need to be uploaded to the cloud. Any file made in this folder should be able to be deleted
and recreated without affecting the larger framework. Right now this folder will be within the
wWorkspace folder, but it is helpful to have the location as a seperate entity to make it possible
in the future to have a smart implementation of a cache of artifacts that can be shared across multiple
workspaces.



TODO: Make a custom type that denotes a path is a relative path to the workspace. 
This could be a pydantic type that way the constraints are validated at runtime.
This is a important part to improve the DX of the working with library. 
"""

import os
from pydantic import DirectoryPath
from pydantic.types import FilePath

from core.constructs.workspace import Workspace

def get_relative_to_workspace_path(fullpath: FilePath) -> str:
    """Generate a relative path starting from the base path of the workspace. 
    
    Note that the given path must be a descendent of the workspace path. 

    Args:
        fullpath (str): Path to convert to a relative path

    Returns:
        str: Relative to the Workspace path
    """

    base_path = Workspace.instance().settings.BASE_PATH
    return os.path.relpath(fullpath, start=base_path)


def get_relative_to_intermediate_path(fullpath: str) -> str:
    """
    Generate a relative path starting from the intermediate directory of the workspace. The given path must be a descendent of the intermediate directory. 

    Args:
        fullpath (str): Path to convert to a relative path

    Returns:
        str: Relative to the intermediate path path
    """
    intermediate_path = Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION
    return os.path.relpath(
        fullpath, start=intermediate_path
    )


def get_full_path_from_workspace_base(relative_path: str) -> FilePath:
    """
    Given a relative path from the workspace path, create the full absolute path on the current filesystem. 

    Args:
        relative_path (str): Relative Path from the workspace path.

    Returns:
        FilePath: Absolute path the file.
    """
    base_path = Workspace.instance().settings.BASE_PATH
    return os.path.join(
        base_path,
        relative_path,
    )


def get_full_path_from_intermediate_base(relative_path: str) -> FilePath:
    """
    Given a relative path from the intermediate folder, create the full absolute path on the current filesystem. 

    Args:
        relative_path (str): Relative Path from the workspace path.

    Returns:
        FilePath: Absolute path the file.
    """
    intermediate_path = Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION
    return os.path.join(
        intermediate_path, relative_path
    )


def is_in_workspace(full_path: str) -> bool:
    """Check if a given file is within the workspace filespace

    Args:
        full_path (str): [description]

    Returns:
        bool: [description]
    """
    base_path = Workspace.instance().settings.BASE_PATH
    return os.path.commonprefix(
        [full_path, base_path]
    ) == base_path


def is_in_intermediate(full_path: str) -> bool:
    """Check if a given file is within the intermediate filespace

    Args:
        full_path (str): [description]

    Returns:
        bool: [description]
    """
    intermediate_path = Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION
    return os.path.commonprefix(
        [full_path, intermediate_path]
    ) == intermediate_path


def get_workspace_path() -> DirectoryPath:
    """Get the Workspace path

    Returns:
        DirectoryPath: Path to workspace directory
    """
    base_path = Workspace.instance().settings.BASE_PATH
    return base_path


def get_intermediate_path() -> DirectoryPath:
    """Get the Intermediate path

    Returns:
        DirectoryPath: Path to intermediate directory
    """
    intermediate_path = Workspace.instance().settings.INTERMEDIATE_FOLDER_LOCATION
    return intermediate_path


def create_path(startingpath, fullpath) -> DirectoryPath:
    """This functions takes a starting path and list of child dir and makes them all

    
    Example

        create_path("./basedir", ["sub1", "sub2"])

        creates:\n
            ./basedir/sub1/\n
            ./basedir/sub1/sub2\n

    Returns:
        DirectoryPath: the final path
    
    """

    if not os.path.isdir(startingpath):
        raise Exception

    intermediate_path = startingpath

    for p in fullpath:
        if not os.path.isdir(os.path.join(intermediate_path, p)):
            os.mkdir(os.path.join(intermediate_path, p))

        intermediate_path = os.path.join(intermediate_path, p)

    return intermediate_path



import os
import setuptools


# Main directory of the project
MAIN_DIR = os.path.dirname(os.getcwd())
print(__file__)
print(MAIN_DIR)
# The text of the README file
README = open(os.path.join(MAIN_DIR, "README.md")).read()


setuptools.setup(
    name='cdev',
    version='0.0.4',
    scripts=['./cdev/scripts/cdev', './core/scripts/cdev_core' ],
    description='CLI for cdev sdk',
    long_description=README,
    long_description_content_type="text/markdown",
    author='CDEV LLC',
    author_email="daniel@cdevframework.com",
    license="Clear BSD",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "boto3",
        "botocore",
        "certifi",
        "charset-normalizer",
        "colorama",
        "commonmark",
        "decorator",
        "docker",
        "gelidum",
        "idna",
        "jmespath",
        "networkx",
        "Parsley",
        "pydantic",
        "Pygments",
        "python-dateutil",
        "python-json-logger",
        "requests",
        "rich",
        "s3transfer",
        "six",
        "sortedcontainers",
        "urllib3",
        "watchdog",
        "websocket-client",
    ],
    python_requires='>=3.7'
)



"""Utilities for generating logs of framework executions

Note that logging should work in a duality within the framework. Meaning, that all functions within the framework
should use the same logging utilities, but how they are provided will differ. All files will have a module
level definition of a default logger, this will allow the utilities to log when called as simple functions
through normal python execution. Within the framework, all executions will be wrapped in a global namespace
that has a logger, therefor when the functions are called as part of the framework, it will be this logger 
that is used.

This is important in mainting that utility functions for the framework are not bound to used within the framework,
but it also allows there to be dynamic functionality of how logs will be proccessed and displayed to the user. 

"""
import logging.config
import logging
import os
from typing import Dict, Union


def _create_logging_settings(log_level: str) -> Dict:
    return {
        "version": 1,
        "filename": ".cdev/logs/userlogs",
        "formatters": {
            "jsonFormatter": {
                "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": "%(asctime)s %(name)s %(levelname)s %(message)s",
            },
            "simpleFormatter": {
                "format": "%(asctime)s %(name)s - %(levelname)s: %(message)s"
            },
            "richFormatter": {"format": "%(name)s - %(levelname)s: %(message)s"},
        },
        "handlers": {
            "fileHandler": {
                "class": "logging.FileHandler",
                "level": log_level,
                "formatter": "jsonFormatter",
                "filename": ".cdev/logs/userlogs",
            },
            "simpleHandler": {
                "class": "logging.StreamHandler",
                "level": log_level,
                "formatter": "simpleFormatter",
            },
            "richHandler": {
                "class": "rich.logging.RichHandler",
                "level": log_level,
                "formatter": "richFormatter",
                "markup": True,
                "show_path": False,
                "show_time": False,
                "rich_tracebacks": True,
            },
        },
        "loggers": {
            "core_file": {
                "level": log_level, 
                "handlers": ["fileHandler"], \
                "propagate": False
            },
            "core_rich": {
                "level": log_level,
                "handlers": ["richHandler"],
                "propagate": False,
            },
            "core_simple": {
                "level": log_level,
                "handlers": ["simpleHandler"],
                "propagate": False,
            },
        },
        "root": {"level": "ERROR", "handlers": ["simpleHandler"]},
        "disable_existing_loggers": False,
    }


class cdev_logger:
    """Wrapper around pythons basic logger object to provide a layer of flexibility for logging. 
    
    Specifically, this will add the ability to always process logs into a JSON file for debugging errors, while
    also allowing the user to provide a flag to set a logging level to display to the user.

    It will implement the functions of a Logger obj
    https://docs.python.org/3/library/logging.html#logging.Logger.propagate


    Note that when formating data into a log message, it must use '%' formatted strings instead of f
    strings. This is allows am optimization used by the logging library to not allocate strings unless the
    log needs to be evaluated. Within the context of a framework, these allocations can add up. This should 
    be enforced with code styling at the PR level. 

    Also note the optimization around formmatted logs. 

    For more info read https://docs.python.org/3/howto/logging.html#optimization
    """

    def __init__(self, is_rich_formatted=True, show_logs: bool=False, logging_level: Union[str,int] = "ERROR") -> None:

        self.is_rich_formatted = is_rich_formatted
        self.show_logs = show_logs
        
        log_info = _create_logging_settings(logging_level)

        fp = os.path.join(os.getcwd(), log_info.get('filename'))
    

        if not os.path.isfile(fp):
            if not os.path.isdir(os.path.dirname(os.path.dirname(fp))):
                os.mkdir(os.path.dirname(os.path.dirname(fp)))

            if not os.path.isdir(os.path.dirname(fp)):
                os.mkdir(os.path.dirname(fp))

            with open(fp, "a"):
                os.utime(fp, None)

        logging.config.dictConfig(log_info)
        self._json_logger = logging.getLogger("core_file")
        self._simple_logger = logging.getLogger("core_simple")
        self._rich_logger = logging.getLogger("core_rich")

    def _write_log(self, *args, func_name: str, original_msg: str, formatted_msg: str = "",  **kw_args):
        # Always write a log to the json file
        getattr(self._json_logger, func_name)(original_msg, *args, **kw_args)
        if self.show_logs:
            # We need to write logs to console
            # Either write a plain log or rich formatted log to the console
            if not self.is_rich_formatted:
                getattr(self._simple_logger, func_name)(original_msg, *args, **kw_args)
            else:
                getattr(self._rich_logger, func_name)(formatted_msg, *args, **kw_args)


    def debug(self, msg, *args, **kw_args):
        if self.is_rich_formatted:
            self._write_log(
                *args,
                func_name="debug", 
                original_msg=msg, 
                formatted_msg=f"[bold blue]{msg}",
                **kw_args
            )

        else:
            self._write_log(
                *args,
                func_name="debug", 
                original_msg=msg, 
                **kw_args
            )


    def info(self, msg, *args, **kw_args):
        if self.is_rich_formatted:
            self._write_log(
                *args,
                func_name="info", 
                original_msg=msg, 
                formatted_msg=f"[bold blue]{msg}",
                **kw_args
            )

        else:
            self._write_log(
                *args,
                func_name="info", 
                original_msg=msg, 
                **kw_args
            )


    def warning(self, msg, *args, **kw_args):
        if self.is_rich_formatted:
            self._write_log(
                *args,
                func_name="warning", 
                original_msg=msg, 
                formatted_msg=f"[bold yellow blink]{msg}",
                **kw_args
            )

        else:
            self._write_log(
                *args,
                func_name="warning", 
                original_msg=msg, 
                **kw_args
            )
            

    def error(self, msg, *args, **kw_args):
        if self.is_rich_formatted:
            self._write_log(
                *args,
                func_name="error",
                original_msg=msg,
                formatted_msg=f"[bold red blink]:cross_mark: :cross_mark: :cross_mark: {msg} :cross_mark: :cross_mark: :cross_mark:",
                **kw_args
            )

        else:
            self._write_log(
                *args,
                func_name="error",
                original_msg=msg,
                **kw_args
            )

    def exception(self, msg):
        self._write_log("exception", msg, msg)


class global_log_container:
    def __init__(self, logger: cdev_logger) -> None:
        self._logger = logger

    @property
    def is_rich_formatted(self) -> bool:
        return self._logger.is_rich_formatted

    @property
    def show_logs(self) -> bool:
        return self._logger.show_logs

    def debug(self, msg, *args, **kw_args):
        self._logger.debug(msg, *args, **kw_args)

    def info(self, msg, *args, **kw_args):
        self._logger.info(msg, *args, **kw_args)

    def warning(self, msg, *args, **kw_args):
        self._logger.warning(msg, *args, **kw_args)
        
    def error(self, msg, *args, **kw_args):
        self._logger.error(msg, *args, **kw_args)

    def exception(self, msg):
        self._logger.exception(msg)
        

log = global_log_container(cdev_logger())



def set_global_logger(new_logger: cdev_logger):
    global log

    log._logger = new_logger






import math
import os
from time import sleep
from typing import Any, Dict, FrozenSet, Union, List, Tuple
from uuid import uuid4

from core.constructs.resource import Resource_Difference, Resource_Change_Type
from core.constructs.output_manager import OutputTask
from core.constructs.models import frozendict

from core.default.resources.simple import xlambda as simple_xlambda
from core.default.resources.simple.iam import permission_arn_model, permission_model

from core.utils import paths as core_paths, hasher
from core.utils.logger import log
from core.utils.platforms import lambda_python_environment



from .. import aws_client 
from boto3.s3.transfer import TransferConfig


#from .lambda_event_deployer import EVENT_TO_HANDLERS
from .role_deployer import (
    create_role_with_permissions,
    delete_role_and_permissions,
    add_policy,
    delete_policy,
    detach_policy,
)


from .event_deployer import (
    EVENT_TO_HANDLERS
)


python_environment_to_aws_params: Dict[lambda_python_environment, Tuple] = {
    lambda_python_environment.py37: ('python3.7', 'x86_64'),
    lambda_python_environment.py38_x86_64: ('python3.8', 'x86_64'),
    lambda_python_environment.py38_arm64: ('python3.8', 'arm64'),
    lambda_python_environment.py39_x86_64: ('python3.9', 'x86_64'),
    lambda_python_environment.py39_arm64: ('python3.9', 'arm64'),
    lambda_python_environment.py3_x86_64: ('python3.9', 'x86_64'),
    lambda_python_environment.py3_arm64: ('python3.9', ' arm64'),
}


AssumeRolePolicyDocumentJSON = """{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Effect": "Allow",
      "Principal": {
        "Service": "lambda.amazonaws.com"
      },
      "Action": "sts:AssumeRole"
    }
  ]
}"""


BUCKET = "cdev-demo-project-artifacts"

###########################
###### Main Handlers
###########################
def _create_simple_lambda(
        transaction_token: str, 
        namespace_token: str, 
        resource: simple_xlambda.simple_function_model, 
        output_task: OutputTask
    ) -> Dict:
    # Steps for creating a deployed lambda function
    # 1. Create IAM Role with needed permissions (note this is first to give aws more time to create the role in all regions and be available for use)
    # 2. Upload the artifact to S3 as the archive location
    # 3. Upload dependencies if needed
    # 4. Create the function
    # 5. Create any integrations that are need based on Events passed in
    log.debug("Create lambda %s", resource)
    full_namespace_suffix = hasher.hash_list([namespace_token, str(uuid4())])

    function_name = f"cdev_function_{full_namespace_suffix}"
    output_task.update(
        comment=f"Creating lambda function resources for lambda {function_name}"
    )

    final_info = {
        'function_name': function_name
    }

    output_task.update(
        comment=f"Creating IAM Role w/ permissions {resource.permissions}"
    )
    # Step 1
    role_name = f"role_{function_name}"
    role_arn, permission_info = create_role_with_permissions(role_name, resource.permissions, AssumeRolePolicyDocumentJSON)

    output_task.update(
        comment=f"Create role for lambda function {resource.name}"
    )

    final_info["role_id"] = role_arn
    final_info["role_name"] = role_name
    final_info["permissions"] = permission_info

    # Step 2
    output_task.update(
        comment=f"Uploading code for lambda function {resource.name}"
    )

    keyname = _upload_s3_code_artifact(function_name, resource)
    

    final_info["artifact_bucket"] = BUCKET
    final_info["artifact_key"] = keyname

    
    # Step 4
    # TODO
    output_task.update(
        comment=f"[blink]Waiting for role to finish creating (~10s)[/blink]"
    )
    sleep(10)
    # ughhhh add a retry wrapper because it takes time to generate the IAM roles across all regions so we need to wait a few seconds to create this
    
    output_task.update(
        comment=f"Create Lambda function"
    )

    runtime, arch = python_environment_to_aws_params.get(resource.platform)
    
    lambda_function_args = {
        "FunctionName": function_name,
        "Runtime": runtime,
        "Architectures": [arch],
        "Role": role_arn,
        "Handler": resource.configuration.handler,
        "Code": {"S3Bucket": BUCKET, "S3Key": keyname},
        "Environment": {
            "Variables":resource.configuration.environment_variables._d
        }
        if resource.configuration.environment_variables
        else {},
        "Layers": list(resource.external_dependencies)
        if resource.external_dependencies
        else [],
    }
    log.debug("lambda configuration %s", lambda_function_args)

    lambda_function_rv = aws_client.run_client_function(
        "lambda", "create_function", lambda_function_args
    )
    final_info["layers"] = resource.external_dependencies
    final_info["cloud_id"] = lambda_function_rv.get("FunctionArn")

    # Step 5
    
    if resource.events:
        available_event_handlers = EVENT_TO_HANDLERS.get(simple_xlambda.RUUID)
        function_cloud_id = final_info.get("cloud_id")

        event_to_output = {}
        for event in resource.events:

            if not event.originating_resource_type in available_event_handlers:
                raise Exception(f'No handlers for {event.originating_resource_type} to {simple_xlambda.RUUID} events')            


            output = EVENT_TO_HANDLERS.get(simple_xlambda.RUUID).get(event.originating_resource_type).get("CREATE")(
                event, function_cloud_id
            )


            event_to_output[event.hash] = output
            

        final_info["events"] = event_to_output

    return final_info


def _remove_simple_lambda(
    transaction_token: str,  
    previous_resource: simple_xlambda.simple_function_model, 
    previous_output: Dict,
    output_task: OutputTask
) -> bool:
    # Steps:
    # Remove and event that is on the function to make sure resources are properly cleaned up
    # Remove the actual function
    # Remove the role associated with the function
    
    cloud_id = previous_output.get('cloud_id')


    if previous_resource.events:
        output_task.update(
            comment=f"Removing Events"
        )
        event_hashes_to_events = {
            x.get('hash'):x for x in [dict(y) for y in previous_resource.events]
        }


        for event_id, event_output in previous_output.get('events').items():        
            originating_resource_type = event_hashes_to_events.get(event_id).get('originating_resource_type')

            EVENT_TO_HANDLERS.get(simple_xlambda.RUUID).get(originating_resource_type).get("REMOVE")(
                event_output,  cloud_id
            )
        

    output_task.update(
        comment=f"Deleting function resource for {cloud_id}"
    )


    aws_client.run_client_function(
        "lambda", "delete_function", {"FunctionName": cloud_id}
    )

    role_arn = previous_output.get("role_id")
    role_name = previous_output.get("role_name")
    permissions =  previous_output.get("permissions")

    delete_role_and_permissions(role_name, permissions)

    output_task.update(
        comment=f"Deleting permissions for the resource ({cloud_id})"
    )

   
    output_task.update(
        comment=f"Removed resources for lambda {cloud_id}"
    )


def _update_simple_lambda(
    transaction_token: str, 
    namespace_token: str,
    previous_resource: simple_xlambda.simple_function_model,
    new_resource: simple_xlambda.simple_function_model,
    previous_output: Dict,
    output_task: OutputTask
) -> bool:
    # Updates can be of:
    # Update source code or dependencies
    # Update configuration
    # Update events

    
    output_task.update(
        comment=f"Updating lambda function {new_resource.name}"
    )

    cloud_id = previous_output['cloud_id']

    did_update_permission = False
    did_update_src_code = False

    mutable_previous_output = dict(previous_output)
    
    # TODO all configurations
    if not previous_resource.configuration == new_resource.configuration:
        if (
            not previous_resource.configuration.environment_variables
            == new_resource.configuration.environment_variables
        ):
            output_task.update(
                comment=f"Updating Environment Variables"
            )
            
            aws_client.run_client_function(
                "lambda",
                "update_function_configuration",
                {
                    "FunctionName": cloud_id,
                    "Environment": {
                        "Variables":new_resource.configuration.environment_variables._d
                    }
                },
            )

    if not previous_resource.permissions == new_resource.permissions:
        output_task.update(
            comment=f"Updating Policies"
        )
        permission_output: frozenset[Dict] = previous_output.get("permissions")
        role_name_output = previous_output.get("role_name")

        remove_permissions = previous_resource.permissions.difference(new_resource.permissions)
        create_permissions = new_resource.permissions.difference(previous_resource.permissions)


        for permission in create_permissions:
            
            rv = frozendict(add_policy(role_name_output, permission))
            tmp = list(permission_output)
            tmp.append(rv)
            permission_output = frozenset(tmp)

        for permission in remove_permissions:

            previous_permission_output_list = [x for x in permission_output if x.get('hash') == permission.hash]

            if len(previous_permission_output_list) > 1:
                raise Exception

            previous_permission_output = previous_permission_output_list[0]
            
            delete_policy(
                role_name_output,
                previous_permission_output
            )

            permission_output = frozenset([x for x in permission_output if not x.get('hash') == permission.hash])

        mutable_previous_output['permissions'] = permission_output
        
        did_update_permission = True


    if not previous_resource.src_code_hash == new_resource.src_code_hash:
        sleep(3)
        output_task.update(
            comment=f"Update Source Code"
        )

        keyname = _upload_s3_code_artifact(previous_output.get('function_name'), new_resource)
        
        aws_client.run_client_function(
            "lambda",
            "update_function_code",
            {
                "FunctionName": cloud_id,
                "S3Key": keyname,
                "S3Bucket": BUCKET,
                "Publish": True,
            },
        )

        mutable_previous_output["artifact_key"] = keyname
        did_update_src_code = True


    if (
        not previous_resource.external_dependencies
        == new_resource.external_dependencies
    ):
        if did_update_src_code:
            output_task.update(
                comment=f"[blink]Waiting for src code update to complete. ~5s[/blink]"
            )
            sleep(5)


        output_task.update(
            comment=f"Update Dependencies"
        )
        
        remove_dependencies = previous_resource.external_dependencies.difference(new_resource.external_dependencies)
        create_dependencies = new_resource.external_dependencies.difference(previous_resource.external_dependencies)
        
        previous_dependency_output: List =  list(mutable_previous_output.get("layers"))

        for dependency in create_dependencies:
            previous_dependency_output.append(dependency)

        for dependency in remove_dependencies:
            previous_dependency_output.remove(dependency)

        aws_client.run_client_function(
            "lambda",
            "update_function_configuration",
            {
                "FunctionName": cloud_id,
                "Layers": previous_dependency_output
            },
        )

        mutable_previous_output['layers'] = previous_dependency_output

    if not previous_resource.events == new_resource.events:
        if did_update_permission:
            # Wait because the updated permission can effect if the event can be bound. 
            output_task.update(
                comment=f"[blink]Waiting for policies to complete. ~10s[/blink]"
            )
            sleep(10)

        output_task.update(
                comment=f"Updating Events"
            )

        create_events = new_resource.events.difference(previous_resource.events)
        remove_events = previous_resource.events.difference(new_resource.events)

        available_event_handlers = EVENT_TO_HANDLERS.get(simple_xlambda.RUUID)

        previous_event_output: dict = dict(mutable_previous_output.get('events')) if mutable_previous_output.get('events') else {}

       
        for _event in create_events:
            if not _event.originating_resource_type in available_event_handlers:
                raise Exception(f'No handlers for {_event.originating_resource_type} to {simple_xlambda.RUUID} events')            


            output_task.update(
                comment=f"Create Event {_event}"
            )

            output = EVENT_TO_HANDLERS.get(simple_xlambda.RUUID).get(_event.originating_resource_type).get("CREATE")(
                _event, cloud_id
            )


            previous_event_output[_event.hash] = output

        previous_event_hashes_to_output = {
            hash:output for hash, output in [ (y.hash, dict(previous_event_output.get(y.hash))) for y in remove_events if previous_event_output.get(y.hash)]
        }

        previous_hashes_to_events = {
            x.get('hash'):x for x in [dict(y) for y in remove_events]
        }

        for event_id, event_output in previous_event_hashes_to_output.items():        
            originating_resource_type = previous_hashes_to_events.get(event_id).get('originating_resource_type')

            output_task.update(
                comment=f"Delete Event {event_id}"
            )

            EVENT_TO_HANDLERS.get(simple_xlambda.RUUID).get(originating_resource_type).get("REMOVE")(
                event_output,  cloud_id
            )

            previous_event_output.pop(event_id)

        mutable_previous_output['events'] = previous_event_output
        
    
    return mutable_previous_output






###############################
###### Upload Artifacts to S3
###############################

def _upload_s3_code_artifact(
    function_name: str, 
    resource: simple_xlambda.simple_function_model,
) -> str:
    # Takes in a resource and create an s3 artifact that can be use as src code for lambda deployment
    keyname = function_name + f"-{resource.hash}" + ".zip"
    # original_zipname = resource.configuration.Handler.split(".")[0] + ".zip"
    zip_location =  core_paths.get_full_path_from_workspace_base(resource.filepath)

    if not os.path.isfile(zip_location):
        # TODO better exception
        raise Exception

    with open(zip_location, "rb") as fh:
        object_args = {
            "Bucket": BUCKET,
            "Key": keyname,
            "Body": fh.read(),
        }
        aws_client.run_client_function("s3", "put_object", object_args)

    return keyname

def _upload_s3_dependency(
    dependency: simple_xlambda.dependency_layer_model,
    output_task: OutputTask
) -> str:
    # Takes in a resource and create an s3 artifact that can be use as src code for lambda deployment
    keyname = f"{dependency.name}-{dependency.hash}.zip"

    zip_location = core_paths.get_full_path_from_workspace_base(dependency.artifact_path)

    total_bytes =  os.path.getsize(zip_location)

    if not os.path.isfile(zip_location):
        # TODO better exception
        raise Exception

    config = TransferConfig(
        multipart_threshold=1024 * 25,
        max_concurrency=10,
        multipart_chunksize=1024 * 25,
        use_threads=True,
    )

    

    def update_progress_bar(x):
        advance_amount = (x/total_bytes) * 4
        output_task.update(advance=advance_amount, comment='[blink]Uploading Package[/blink]')


    object_args = {
        "Bucket": BUCKET,
        "Key": keyname,
        "Filename": zip_location,
        "Callback": update_progress_bar,
        "Config": config,
    }
    aws_client.run_client_function("s3", "upload_file", object_args)

    output_task.update( comment='Uploaded Package')

    return keyname


#######################
##### Main Entry Point
#######################

def handle_simple_lambda_function_deployment( 
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    try:
        log.debug("Calling lambda mapper")
        if resource_diff.action_type == Resource_Change_Type.CREATE:
            return _create_simple_lambda(
                transaction_token,
                namespace_token,
                resource_diff.new_resource,  
                output_task
            )
        elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:

            return _update_simple_lambda(
                transaction_token,
                namespace_token,
                simple_xlambda.simple_function_model(**resource_diff.previous_resource.dict()),
                resource_diff.new_resource,
                previous_output,
                output_task
            )
        elif resource_diff.action_type == Resource_Change_Type.DELETE:

            return _remove_simple_lambda(
                transaction_token,
                resource_diff.previous_resource,
                previous_output,
                output_task
            )

    except Exception as e:
        print(e)
        raise Exception("COULD NOT DEPLOY")

###################################
##### Layers
###################################

def _create_simple_layer( 
        transaction_token: str, 
        namespace_token: str, 
        resource: simple_xlambda.dependency_layer_model, 
        output_task: OutputTask
    ):

    output_task.update(
        comment=f"Creating dependencies for lambda function {resource.name}"
    )

    key_name = _upload_s3_dependency(resource, output_task)

    # key name will always include .zip so remove that part and change '-' into '_'
    layer_name = key_name.replace("-", "_")[:-4].replace(".","")
    dependency_rv = aws_client.run_client_function(
        "lambda",
        "publish_layer_version",
        {
            "Content": {"S3Bucket": BUCKET, "S3Key": key_name},
            "LayerName": f"{layer_name}_{namespace_token[:10]}",
            "CompatibleRuntimes": [
                "python3.7",
                "python3.8",
                "python3.9"
            ],
        },
    )

    return {
        "cloud_id": f"{dependency_rv.get('LayerArn')}:{dependency_rv.get('Version')}",
        "arn": dependency_rv.get("LayerArn"),
        "version": dependency_rv.get('Version')
    }

def _update_simple_layer(
        transaction_token: str, 
        namespace_token: str,
        previous_resource: simple_xlambda.dependency_layer_model,
        new_resource: simple_xlambda.dependency_layer_model,
        previous_output: Dict,
        output_task: OutputTask
    ):
    return _create_simple_layer(
        transaction_token,
        namespace_token,
        new_resource,
        output_task
    )
    
def _remove_simple_layer(
        transaction_token,
        resource,
        previous_output,
        output_task
    ):
        aws_client.run_client_function(
            "lambda",
            "delete_layer_version",
            {
                "LayerName": previous_output.get("arn"),
                "VersionNumber": previous_output.get("version"),
            },
        )



def handle_simple_layer_deployment( 
        transaction_token: str, 
        namespace_token: str, 
        resource_diff: Resource_Difference, 
        previous_output: Dict[str, Any],
        output_task: OutputTask
    ) -> Dict:
    try:
        if resource_diff.action_type == Resource_Change_Type.CREATE:
            return _create_simple_layer(
                transaction_token,
                namespace_token,
                resource_diff.new_resource,  
                output_task
            )
        elif resource_diff.action_type == Resource_Change_Type.UPDATE_IDENTITY:

            return _update_simple_layer(
                transaction_token,
                namespace_token,
                simple_xlambda.dependency_layer_model(**resource_diff.previous_resource.dict()),
                resource_diff.new_resource,
                previous_output,
                output_task
            )

        elif resource_diff.action_type == Resource_Change_Type.DELETE:
            return _remove_simple_layer(
                transaction_token,
                resource_diff.previous_resource,
                previous_output,
                output_task
            )

    except Exception as e:
        print(e)
        raise Exception("COULD NOT DEPLOY")



"""Set of constructs for creating HTTP APIs

"""
from enum import Enum
from typing import Any, Dict, FrozenSet, List, Optional

from core.constructs.models import ImmutableModel
from core.constructs.cloud_output import Cloud_Output_Mapping, Cloud_Output_Sequence, Cloud_Output_Str, OutputType

from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs
from core.constructs.types import cdev_str_model
from core.utils import hasher

from core.default.resources.simple import events


RUUID = "cdev::simple::api"

########################
##### Authorizer
########################

class authorizer_model(ImmutableModel):
    """Model representing JWT authorizer information"""
    name: str
    issuer_url: str
    audience: str


class Authorizer():
    def __init__(self, name: str, issuer_url: str, audience: str) -> None:
        """JWT authorizer information.

        Deployed Apis can validates the JWTs that clients submit with API requests to allow or deny 
        requests based on token validation, and optionally, scopes in the token. You can provide authorization
        from OpenID Connect (OIDC) and OAuth 2.0 frameworks. 

        For examples on how to integrate with service like Auth0 visit our 
        <a href="/docs/examples/user-authentication"> documentation </a>.


        Args:
            name (str): Name of the authorizer. Used as a unique identifier within the attached Api.
            issuer_url (str): The base domain of the identity provider that issues JSON Web Tokens.
            audience (str): The intended recipients of the JWT. A valid JWT must provide an aud that matches at this value.
        """
        self.name = name
        self.issuer_url = issuer_url
        self.audience = audience


    def render(self) -> authorizer_model:
        return authorizer_model(
            name=self.name,
            issuer_url=self.issuer_url,
            audience=self.audience,
        )

    def hash(self) -> str:
        return hasher.hash_list([
            self.name,
            self.issuer_url,
            self.audience
        ])



########################
##### Route
########################
class route_verb(str,Enum):
    """Route Verbs"""
    GET = "GET"
    PUT = "PUT"
    POST = "POST"
    DELETE = "DELETE"
    ANY = "ANY"


class route_model(ImmutableModel):
    """Model to represent a route of an API"""
    path: str
    verb: str
    additional_scopes: Optional[FrozenSet[str]]
    authorizer_name: Optional[str]

    def __init__(__pydantic_self__, **data: Any) -> None:
        """"""
        super().__init__(**data)


class Route():
    def __init__(self, api_name: str, path: str, verb: route_verb, authorizer_name: str = None, additional_scopes: List[str] = []) -> None:
        """Construct for representing a route that is apart of an HTTP API 

        Args:
            api_name (str): Cdev name of the API this route is apart of
            path (str): Path of the route.
            verb (route_verb): Verb that this route handles
            authorizer_name (str): The `name` of the authorizer to use for this route.
            additional_scopes (List[str]): Set of scopes that must be present in an authorization JWT.
        """
        self.api_name = api_name
        self.path = path
        self.verb = verb
        self.additional_scopes = additional_scopes
        self.authorizer_name = authorizer_name
       
    def hash(self) -> str:
        return hasher.hash_list([
            self.path,
            self.verb,
            hasher.hash_list(self.additional_scopes),
            self.authorizer_name
        ])

    def render(self) -> route_model:
        return route_model(
            path=self.path,
            verb=self.verb,
            additional_scopes=frozenset(self.additional_scopes),
            authorizer_name=self.authorizer_name
        )

    def event(self) -> 'RouteEvent':
        """Generate an Event Construct that other resources can bind to. 

        Returns:
            `RouteEvent`
        """
        return RouteEvent(
            resource_name=self.api_name,
            path=self.path,
            verb=self.verb
        )

class route_event_model(events.event_model):
    """Model to represent an event for a given route"""
    path: str
    verb: route_verb
    api_id: cdev_str_model
    route_id: cdev_str_model

    def __init__(__pydantic_self__, **data: Any) -> None:
        """"""
        super().__init__(**data)


class RouteEvent():
    """Construct for representing a route that is apart of an HTTP API."""

    def __init__(self, resource_name: str, path: str, verb: route_verb) -> None:
        """
        Args:
            resource_name (str): Cdev Name of the API this event is generated from
            path (str): Path of the route
            verb (`route_verb`): Verb of the route

        This construct should be generated with the `Route.event` method.
        """
        self.resource_name = resource_name
        self.path = path
        self.verb = verb
        self.api_id = Cloud_Output_Str(
                name=resource_name, 
                ruuid=RUUID, 
                key='cloud_id',
                type=OutputType.RESOURCE 
            )
        self.route_id = Cloud_Output_Mapping[Cloud_Output_Str](
                name=resource_name, 
                ruuid=RUUID, 
                key='endpoints',
                type=OutputType.RESOURCE,
                _member_class=Cloud_Output_Str
            )[f'{self.path} {self.verb}']

    def hash(self) -> str:
        return hasher.hash_list([self.resource_name, self.path, self.verb])

    def render(self) -> route_event_model:
        return route_event_model(
            originating_resource_name=self.resource_name,
            originating_resource_type=RUUID,
            hash=self.hash(),
            path=self.path,
            verb=self.verb,
            api_id=self.api_id.render(),
            route_id=self.route_id.render()
        )

########################
##### Output
########################
class ApiOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after an API has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)
    
    @property
    def endpoint(self) -> Cloud_Output_Str:
        """The base url for the created API

        Returns:
            `core.constructs.cloud_output.Cloud_Output_Str`
        """
        return Cloud_Output_Str(
            name=self._name,
            ruuid=RUUID,
            key='endpoint',
            type=self.OUTPUT_TYPE
        )

    @endpoint.setter
    def endpoint(self, value: Any):
        raise Exception


    @property
    def endpoints(self) -> Cloud_Output_Mapping[Cloud_Output_Str]:
        """The created routes for the API.

        Values are stored as a mapping of the (path,verb) to cloud route_id.  
        
        ex:
        ```
            {
                '/hello_world [GET]': <route_id>
            }
        ```

        Returns:
            `core.constructs.cloud_output.Cloud_Output_Mapping`
        """
        return Cloud_Output_Mapping[Cloud_Output_Str](
            name=self._name,
            ruuid=RUUID,
            key='endpoints',
            type=self.OUTPUT_TYPE,
            _member_class=Cloud_Output_Str
        )

    @endpoints.setter
    def endpoints(self, value: Any):
        raise Exception


########################
##### Api
########################
class simple_api_model(ResourceModel):
    """Model for representing a desired HTTP API"""
    
    routes: FrozenSet[route_model]
    allow_cors: bool
    authorizers: Optional[FrozenSet[authorizer_model]]

    def __init__(self, **data: Any) -> None:
        """"""
        super().__init__(**data)


class Api(Resource):

    @update_hash
    def __init__(
        self, cdev_name: str, allow_cors: bool = True, authorizers: List[Authorizer] = [], default_authorizer: str= None, nonce: str = "",
    ):
        """Create a HTTP API.

        Args:
            cdev_name (str): Name for the resource.
            allow_cors (bool): Allow Cross Origin Resource Sharing (CORS) on the api.
            authorizers (List[Authorizer]): List of JWT Authorizers for the api. 
            default_authorizer (str): The name of an authorizer to add as the default to all routes. 
            nonce (str): Nonce to make the resource hash unique if there are conflicting resources with same configuration.

        With an HTTP API, you can create different routes that represent different requests to your backend service. Use the 
        `route` method to create these routes then attach them to other resource to handle the requests.

        HTTP Api's can take an authorizer to support JWT authorization for the Api. You can provide a default authorizer to the Api to apply
        to all created routes, or individually set the authorizer for each route.

        <a href="https://code.tutsplus.com/tutorials/a-beginners-guide-to-http-and-rest--net-16340"> More information on HTTP routes</a>

        <a href="/docs/examples/api"> Examples on how to use in Cdev Framework </a>

        <a href="https://docs.aws.amazon.com/apigatewayv2/latest/api-reference/api-reference.html"> Documentation on Deployed Resource in the Cloud</a>

        <a href="https://aws.amazon.com/api-gateway/pricing/"> Details on pricing</a>
        """

        super().__init__(cdev_name, RUUID, nonce)
        
        self._allow_cors = allow_cors
        self._routes: List[Route] = []

        self._authorizers: List[Authorizer] = authorizers

        self._default_authorizer_name = default_authorizer

        self._output = ApiOutput(cdev_name)


    @property
    def output(self) -> ApiOutput:
        """Output generated by the Cloud when this resource is deployed."""
        return self._output

    @property
    def allow_cors(self) -> bool:
        """Allow the API to be accessed from sites other than the generated url.

        [Notes on CORS]
        """
        return self._allow_cors

    @allow_cors.setter
    @update_hash
    def allow_cors(self, value: bool):
        self._allow_cors = value

    @update_hash
    def route(self, path: str, verb: route_verb, additional_scopes: List[str] = [], override_authorizer_name: str = None) -> Route:
        """Create a route for the API.

        Args:
            path (str): The http path of the route created
            verb (`route_verb`): verb for the path
            override_authorizer_name (str): The authorizer for this route. Overriding the default authorizer for the Api. 
        Returns:
            `Route` 

        Generate a `Route` that can be used as a trigger for other resources.

        ```
        route = myApi.route('/hello_world', 'GET')
        ```
        """

        authorizer_name = override_authorizer_name if override_authorizer_name else self._default_authorizer_name
    
        route = Route(
            self.name,
            path, 
            verb,
            authorizer_name,
            additional_scopes
            
        )

        self._routes.append(route)

        return route

    def compute_hash(self):
        self._hash = hasher.hash_list(
            [
                hasher.hash_list([x.hash() for x in self._routes]),
                self.allow_cors, 
                self.nonce,
                self._default_authorizer_name,
                hasher.hash_list([x.hash() for x in self._authorizers])
            ]
        )

    def render(self) -> simple_api_model:
        routes =  frozenset(self._routes)
       
        return simple_api_model(
            ruuid=self.ruuid,
            name=self.name,
            hash=self.hash,
            routes=frozenset([x.render() for x in routes]),
            allow_cors=self.allow_cors,
            authorizers=frozenset([x.render() for x in self._authorizers])
        )



"""Utilities for creating a new resource state

"""

from ..constructs.workspace import Workspace


def create_resource_state_cli(args):

    WORKSPACE = Workspace.instance()

    create_resource_state(WORKSPACE)


def create_resource_state(workspace: Workspace):
    new_uuid = workspace.get_backend().create_resource_state("demo")
    



from cdev.constructs.project import Project
from cdev.cli.logger import set_global_logger_from_cli
from cdev.default.output_manager import CdevOutputManager

from core.commands.run import run_command as core_run_command


def run_command_cli(args):
    config = args[0]
    set_global_logger_from_cli(config.loglevel)

    run_command(args[0])


def run_command(args):
    output_manager = CdevOutputManager()

    myProject = Project.instance()
    
    ws = myProject.get_current_environment().get_workspace()
    
    core_run_command(ws, output_manager, args)


import json
import os

from pathlib import PosixPath, WindowsPath
from pydantic.main import BaseModel
from pydantic.types import DirectoryPath, FilePath

from typing import Dict, List, Any, Tuple
import uuid

from core.constructs.workspace import Workspace


from ..constructs.backend import Backend_Configuration, Backend
from ..constructs.backend_exceptions import *

from ..constructs.components import (
    Component_Change_Type,
    ComponentModel,
    Component_Difference,
)
from ..constructs.resource import (
    Resource_Change_Type,
    Resource_Difference,
    Resource_Reference_Change_Type,
    ResourceModel,
    Resource_Reference_Difference,
    ResourceReferenceModel,
)
from ..constructs.resource_state import Resource_State
from ..utils import file_manager, hasher as cdev_hasher


class LocalBackendError(BackendError):
    pass


class CanNotFindResourceStateFile(LocalBackendError):
    pass


class InvalidResourceStateData(LocalBackendError):
    pass



class SetEncoder(json.JSONEncoder):
    def default(self, obj):
       if isinstance(obj, set):
          return list(obj)
       return json.JSONEncoder.default(self, obj)


class Local_Backend_Configuration(Backend_Configuration):
    def __init__(self, config: Dict) -> None:
        """
        Represents the data needed to create a new cdev workspace:

        Parameters:
            python_module: The name of the python module to load as the backend
            config: configuration option for the backend

        """

        super().__init__(
            **{
                "python_module": "core.default.backend",
                "python_class": "LocalBackend",
                "config": config,
            }
        )


class LocalCentralFile(BaseModel):
    resource_state_locations: Dict[str, str]  # uuid -> file location
    top_level_states: List[str]  # uuid
    resource_state_names: List[str]

    def __init__(
        __pydantic_self__,
        resource_state_locations: Dict[str, str],
        top_level_states: List[str],
        resource_state_names: List[str],
    ) -> None:
        super().__init__(
            **{
                "resource_state_locations": resource_state_locations,
                "top_level_states": top_level_states,
                "resource_state_names": resource_state_names,
            }
        )


class LocalBackend(Backend):
    # Structurally, this implementation will have a central json that can be used as an index into more precise json files. For example, each resource state will be its own json file, but
    # the central file will keep track of each one.
    def __init__(
        self, base_folder: DirectoryPath, central_state_file: FilePath = None
    ) -> None:
        """
        Implementation of a Backend using locally stored json files as the peristent storage medium. This backend should only be used for small project as it does not provide any mechanisms
        to work well when multiple people edit the state. Also this is a single threaded implementation.

        *** For now, we will not use any kind of WAL for make changes to underlying state files, so it can be bad if you kill the process unexpectedly. In the future, it will use some mechanism
        to prevent this. ***

        Arguments:
            base_folder (DirectoryPath): Path to a folder to use for storing local json files. Defaults to cdev setting if not provided.
            central_state_file (FilePath): Path to the central state file. Defaults to cdev setting if not provided.
        """

        
        DEFAULT_CENTRAL_STATE_FILE = os.path.join(base_folder, "local_state.json")


        self.base_folder = base_folder
        self.central_state_file = (
            central_state_file if central_state_file else DEFAULT_CENTRAL_STATE_FILE
        )

        if not os.path.isdir(self.base_folder):
            print(f"ERROR HERE")
            raise Exception

        if not os.path.isfile(self.central_state_file):
            self._central_state = LocalCentralFile({}, [], [])

        else:
            with open(self.central_state_file, "r") as fh:
                self._central_state = LocalCentralFile(**json.load(fh))

    def _write_central_file(self):
        file_manager.safe_json_write(self._central_state.dict(), self.central_state_file)

    def _write_resource_state_file(self, resource_state: Resource_State, fp: FilePath):
        file_manager.safe_json_write(resource_state.dict(), fp)
    

    # Api for working with Resource States
    def create_resource_state(
        self, name: str, parent_resource_state_uuid: str = None
    ) -> str:
        # Create the new resource state
        if name in set(self._central_state.resource_state_names):
            raise ResourceStateAlreadyExists("Creating resource state with taken name")

        resource_state_uuid = str(uuid.uuid4())
        if not parent_resource_state_uuid:
            # Create this as a top level resource state
            new_resource_state = Resource_State(name, resource_state_uuid)
            self._central_state.top_level_states.append(new_resource_state.uuid)

        else:
            new_resource_state = Resource_State(
                name=name,
                uuid=resource_state_uuid,
                components=[],
                parent_uuid=parent_resource_state_uuid,
            )

        filename = os.path.join(
            self.base_folder, f"resource_state_{new_resource_state.uuid}.json"
        )

        self._central_state.resource_state_locations[new_resource_state.uuid] = filename
        self._central_state.resource_state_names.append(new_resource_state.name)

        self._write_central_file()
        self._write_resource_state_file(new_resource_state, filename)

        return new_resource_state.uuid

    def delete_resource_state(self, resource_state_uuid: str):
        if not resource_state_uuid in self._central_state.resource_state_locations:
            raise ResourceStateDoesNotExist(
                f"Resource state: {resource_state_uuid} does not exist"
            )

        resource_state_to_delete = self.get_resource_state(resource_state_uuid)

        if resource_state_to_delete.children:
            # Can not delete resource state with children
            raise ResourceStateNotEmpty("Resource state has child resource states")

        if not len(resource_state_to_delete.components) == 0:
            # Can not delete a non empty resource state
            raise ResourceStateNotEmpty("Resource state has components")

        if resource_state_to_delete.parent_uuid:
            # Need to remove this as child of parent
            pass

        else:
            # if the resource state had no parent, then it was a top level resource state
            self._central_state.top_level_states.remove(resource_state_to_delete.uuid)

        self._central_state.resource_state_names.remove(resource_state_to_delete.name)
        file_location = self._central_state.resource_state_locations.pop(
            resource_state_to_delete.uuid
        )

        self._write_central_file()
        os.remove(file_location)

    def get_resource_state(self, resource_state_uuid: str) -> Resource_State:
        if not resource_state_uuid in self._central_state.resource_state_locations:
            raise ResourceStateDoesNotExist(
                f"Resource State: {resource_state_uuid} does not exists"
            )

        file_location = self._central_state.resource_state_locations.get(
            resource_state_uuid
        )

        if not os.path.isfile(file_location):
            raise CanNotFindResourceStateFile(
                f"Can not find resource state file for {resource_state_uuid}"
            )

        try:
            return file_manager.load_resource_state(file_location)


        except Exception as e:
            raise InvalidResourceStateData(
                f"Ivalid data for Resource State from file {file_location} for resource state {resource_state_uuid}; {e}"
            )

    def get_top_level_resource_states(self) -> List[Resource_State]:
        rv = []
        
        for resource_id in self._central_state.top_level_states:
            # Let any exception from loading a state pass up to caller
            rv.append(self.get_resource_state(resource_id))

        
        return rv

    # Components
    def _create_component(self, resource_state_uuid: str, component_name: str):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if component_name in set(x.name for x in resource_state.components):
            # Cant not have two components of the same name in the same resource state
            raise ComponentAlreadyExists(
                f"Component already exists with name {component_name} in Resource State {resource_state_uuid}"
            )

        # Create the new component
        new_component = ComponentModel(component_name)
        resource_state.components.append(new_component)

        # Create a uuid for the component
        component_uuid = str(uuid.uuid4())
        resource_state.component_name_to_uuid[component_name] = component_uuid

        self._write_resource_state_file(resource_state, resource_state_file_location)

    def _delete_component(self, resource_state_uuid: str, component_name: str):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not component_name in set(x.name for x in resource_state.components):
            # Component of that name does not exists
            raise ComponentDoesNotExist(
                f"Could not find component {component_name} in Resource State {resource_state_uuid}"
            )

        deleteing_component = next(
            x for x in resource_state.components if x.name == component_name
        )

        if not len(deleteing_component.resources) == 0:
            raise ComponentNotEmpty(
                f"Can not delete Component {component_name} in Resource State {resource_state_uuid} because the component is not empty"
            )

        resource_state.components = [
            x for x in resource_state.components if not x.name == component_name
        ]

        
        resource_state.component_name_to_uuid.pop(component_name)

        self._write_resource_state_file(resource_state, resource_state_file_location)


    def _update_component_name(self, resource_state_uuid: str, previous_component_name: str, new_component_name: str):
        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not previous_component_name in set(x.name for x in resource_state.components) or not previous_component_name in resource_state.component_name_to_uuid:
            # Component of that name does not exists
            raise ComponentDoesNotExist(
                f"Could not find component {previous_component_name} in Resource State {resource_state_uuid}"
            )


        if new_component_name in set(x.name for x in resource_state.components) or new_component_name in resource_state.component_name_to_uuid:
            # Cant not have two components of the same name in the same resource state
            raise ComponentAlreadyExists(
                f"Component already exists with name {new_component_name} in Resource State {resource_state_uuid}"
            )

        # get the component to rename
        rename_component = next(
            x for x in resource_state.components if x.name == previous_component_name
        )

        # remove the component from the list of components
        resource_state.components = [
            x for x in resource_state.components if not x.name == previous_component_name
        ]

        # Since ComponentModels are frozen, we can just change the name
        # So we make a dict of the current ComponentModel then change the name and use the dict as input
        # for a new ComponentModel Obj. Then add the new Obj to the resource state
        component_as_dict = rename_component.dict()
        component_as_dict['name'] = new_component_name
        new_component = ComponentModel(**component_as_dict)

        resource_state.components.append(new_component)

        # Update the name to uuid dict
        resource_state.component_name_to_uuid.pop(previous_component_name)
        resource_state.component_name_to_uuid[new_component_name] = new_component


        self._write_resource_state_file(resource_state, resource_state_file_location)

        

    def get_component(
        self, resource_state_uuid: str, component_name: str
    ) -> ComponentModel:

        resource_state = self.get_resource_state(resource_state_uuid)

        if not component_name in set(x.name for x in resource_state.components):
            # Component of that name does not exists
            raise ComponentDoesNotExist(
                f"Can not find Component {component_name} in Resource State {resource_state_uuid}"
            )

        return next(x for x in resource_state.components if x.name == component_name)


    def get_component_uuid(self, resource_state_uuid: str, component_name: str) -> str:
        resource_state = self.get_resource_state(resource_state_uuid)

        if not component_name in resource_state.component_name_to_uuid:
            raise ComponentDoesNotExist(
                f"Can not find Component {component_name} in Resource State {resource_state_uuid}"
            )

        component_uuid = resource_state.component_name_to_uuid.get(component_name)

        return cdev_hasher.hash_list([resource_state_uuid, component_uuid])


    def update_component(
        self, resource_state_uuid: str, component_difference: Component_Difference
    ):
        if component_difference.action_type == Component_Change_Type.CREATE:
            self._create_component(resource_state_uuid , component_difference.new_name)
            return

        elif component_difference.action_type == Component_Change_Type.DELETE:
            self._delete_component(resource_state_uuid, component_difference.previous_name)
            return

        elif component_difference.action_type == Component_Change_Type.UPDATE_IDENTITY:
            return

        elif  component_difference.action_type == Component_Change_Type.UPDATE_NAME:
            self._update_component_name(resource_state_uuid, component_difference.previous_name, component_difference.new_name)

        else:
            raise Exception(f"Component Action type not supported {component_difference.action_type}")


    def create_resource_change_transaction(
        self, resource_state_uuid: str, component_name: str, diff: Resource_Difference
    ) -> Tuple[str, str]:
        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        transaction_token = str(uuid.uuid4())

        ruuid = diff.new_resource.ruuid if diff.new_resource else diff.previous_resource.ruuid

        namespace_token = cdev_hasher.hash_list([resource_state_uuid, self.get_component_uuid(resource_state_uuid, component_name), ruuid])

        resource_state.resource_changes[transaction_token] = (component_name, diff)

        self._write_resource_state_file(resource_state, resource_state_file_location)

        return transaction_token, namespace_token


    def complete_resource_change(
        self,
        resource_state_uuid: str,
        component_name: str,
        diff: Resource_Difference,
        transaction_token: str,
        cloud_output: Dict = None,
        resolved_cloud_information: Dict={}
    ):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        
        if not transaction_token in resource_state.resource_changes:
            
            raise ResourceChangeTransactionDoesNotExist(
                f"Transaction {transaction_token} does not exist in Resource State {resource_state_uuid}"
            )

        component = self.get_component(resource_state_uuid, component_name)
        
        new_component = self._update_component(component, diff, cloud_output, resolved_cloud_information)
        
        
        resource_state.components = [
            x for x in resource_state.components if not x.name == component.name
        ] + [new_component]

        resource_state.resource_changes.pop(transaction_token)

        self._write_resource_state_file(resource_state, resource_state_file_location)
        

    def fail_resource_change(
        self,
        resource_state_uuid: str,
        component_name: str,
        diff: Resource_Difference,
        transaction_token: str,
        failed_state: Dict,
    ):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not transaction_token in resource_state.resource_changes:
            raise ResourceChangeTransactionDoesNotExist(
                f"Transaction {transaction_token} does not exist in Resource State {resource_state_uuid}"
            )

        resource_state.resource_changes.pop(transaction_token)
        resource_state.failed_changes[transaction_token] = (
            component_name,
            diff,
            failed_state,
        )

        self._write_resource_state_file(resource_state, resource_state_file_location)

    def change_failed_state_of_resource_change(
        self, resource_state_uuid: str, transaction_token: str, new_failed_state: Dict
    ):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not transaction_token in resource_state.failed_changes:
            raise ResourceChangeTransactionDoesNotExist(
                f"Transaction {transaction_token} does not exist in Resource State {resource_state_uuid}"
            )

        previous_component_name, previous_diff, _ = resource_state.failed_changes.get(
            transaction_token
        )

        resource_state.failed_changes[transaction_token] = (
            previous_component_name,
            previous_diff,
            new_failed_state,
        )

        self._write_resource_state_file(resource_state, resource_state_file_location)

    def recover_failed_resource_change(
        self,
        resource_state_uuid: str,
        transaction_token: str,
        to_previous_state: bool = True,
        cloud_output: Dict = None,
    ):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not transaction_token in resource_state.failed_changes:
            raise ResourceChangeTransactionDoesNotExist(
                f"Transaction {transaction_token} does not exist in Resource State {resource_state_uuid}"
            )

        component_name, diff, _ = resource_state.failed_changes.pop(transaction_token)

        if not to_previous_state:
            component = self.get_component(component_name)

            new_component = self._update_component(component, diff, cloud_output)

            resource_state.components = [
                x for x in resource_state.components if not x.name == component_name
            ] + [new_component]

        self._write_resource_state_file(resource_state, resource_state_file_location)

    def remove_failed_resource_change(
        self, resource_state_uuid: str, transaction_token: str
    ):

        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        if not transaction_token in resource_state.failed_changes:
            raise ResourceChangeTransactionDoesNotExist(
                f"Transaction {transaction_token} does not exist in Resource State {resource_state_uuid}"
            )

        resource_state.failed_changes.pop(transaction_token)

        self._write_resource_state_file(resource_state, resource_state_file_location)

    def resolve_reference_change(
        self,
        resource_state_uuid: str,
        diff: Resource_Reference_Difference,
    ):
        resource_state = self.get_resource_state(resource_state_uuid)
        resource_state_file_location = self._get_resource_state_file_location(
            resource_state_uuid
        )

        component = self.get_component(resource_state_uuid, diff.originating_component_name)

        _reference_resource_state = resource_state
        if diff.resource_reference.is_in_parent_resource_state:
            if not resource_state.parent_uuid:
                raise ResourceReferenceError(
                    f"Current Resource State {resource_state_uuid} does not have a Parent Resource State to resolve {diff} to"
                )

            _reference_resource_state = self.get_resource_state(
                resource_state.parent_uuid
            )

        try:
            _referenced_component = self.get_component(
                _reference_resource_state.uuid, diff.resource_reference.component_name
            )
        except ComponentDoesNotExist:
            raise ResourceReferenceError(
                f"Resource State {_reference_resource_state.uuid} does not contain component {diff.resource_reference.component_name} for {diff}"
            )

        all_parent_resources = set(
            [f"{x.ruuid}{x.name}" for x in _referenced_component.resources]
        )

        if (
            not f"{diff.resource_reference.ruuid}{diff.resource_reference.name}"
            in all_parent_resources
        ):
            raise ResourceReferenceError(
                f"Could not find resource {diff.resource_reference.ruuid};{diff.resource_reference.name} in parent component"
            )

        if diff.action_type == Resource_Reference_Change_Type.CREATE:
            
            reference_id = (
                f"{diff.resource_reference.ruuid}{diff.resource_reference.name}"
            )

            # resolve the reference by adding a count to the reference counter in the referenced component
            if not reference_id in _referenced_component.external_references:
                _referenced_component.external_references[reference_id] = {"cnt": 1}

            else:
                previous_cnt = _referenced_component.external_references[
                    reference_id
                ].get("cnt")
                _referenced_component.external_references[reference_id] = {
                    "cnt": previous_cnt + 1
                }

            # Add this to the references for this component
            component.references.append(diff.resource_reference)

        elif diff.action_type == Resource_Reference_Change_Type.DELETE:
            
            reference_id = (
                f"{diff.resource_reference.ruuid}{diff.resource_reference.name}"
            )

            # resolve the dereference by subtracting a count to the reference counter in the referenced component
            if not reference_id in _referenced_component.external_references:
                raise ResourceReferenceError(
                    f"Trying to deference resource that does not have reference info"
                )

            else:
                previous_cnt = _referenced_component.external_references[
                    reference_id
                ].get("cnt")
                _referenced_component.external_references[reference_id] = {
                    "cnt": previous_cnt - 1
                }

            if _referenced_component.external_references[reference_id] == 0:
                _referenced_component.external_references.pop(reference_id)

            # Pop this references for this component
            component.references.remove(diff.resource_reference)

        _reference_resource_state.components = [
            x
            for x in resource_state.components
            if not x.name == _referenced_component.name
        ] + [_referenced_component]
        resource_state.components = [
            x for x in resource_state.components if not x.name == component.name
        ] + [component]

        if diff.resource_reference.is_in_parent_resource_state:
            # if the reference is from the parent resource state then we need to update it so that it knows that the reference was resolved
            _reference_resource_state_fp = self._get_resource_state_file_location(
                _reference_resource_state.uuid
            )

            self._write_resource_state_file(
                _reference_resource_state, _reference_resource_state_fp
            )

        self._write_resource_state_file(resource_state, resource_state_file_location)

    # Get resources and cloud output
    def get_resource_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
    ) -> ResourceModel:

        component = self.get_component(resource_state_uuid, component_name)

        resource = next(
            (
                x
                for x in component.resources
                if x.ruuid == resource_type and x.name == resource_name
            ),
            None,
        )

        if not resource:
            raise ResourceDoesNotExist(
                f"Resource {resource_type}::{resource_name} does not exist in Component {component_name} in Resource State {resource_state_uuid}"
            )

        return resource

    def get_resource_by_hash(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_hash: str,
    ) -> ResourceModel:

        component = self.get_component(resource_state_uuid, component_name)

        resource = next(
            (
                x
                for x in component.resources
                if x.ruuid == resource_type and x.hash == resource_hash
            ),
            None,
        )

        if not resource:
            raise ResourceDoesNotExist(
                f"Resource {resource_type}::{resource_hash} does not exist in Component {component_name} in Resource State {resource_state_uuid}"
            )

        return resource

    def get_cloud_output_value_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
        key: str,
    ) -> Any:

        component = self.get_component(resource_state_uuid, component_name)
        resource = self.get_resource_by_name(
            resource_state_uuid, component_name, resource_type, resource_name
        )

        cloud_output_id = self._get_cloud_output_id(resource)

        if not cloud_output_id in component.cloud_output:

            raise CloudOutputDoesNotExist(
                f"Can not find Cloud Output for {resource_type}::{resource_name} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        cloud_output = component.cloud_output.get(cloud_output_id)

        if not cloud_output:
            raise CloudOutputDoesNotExist(
                f"None value for Cloud Output for {resource_type}::{resource_name} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        if not key in cloud_output:
            raise KeyNotInCloudOutput(
                f"Can not find Key {key} in Cloud Output for {resource_type}::{resource_name} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        return cloud_output.get(key)

    def get_cloud_output_value_by_hash(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_hash: str,
        key: str,
    ) -> Any:

        component = self.get_component(resource_state_uuid, component_name)
        resource = self.get_resource_by_hash(
            resource_state_uuid, component_name, resource_type, resource_hash
        )

        cloud_output_id = self._get_cloud_output_id(resource)

        if not cloud_output_id in component.cloud_output:
            raise CloudOutputDoesNotExist(
                f"Can not find Cloud Output for {resource_type}::{resource_hash} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        cloud_output = component.cloud_output.get(cloud_output_id)

        if not cloud_output:
            raise CloudOutputDoesNotExist(
                f"None value for Cloud Output for {resource_type}::{resource_hash} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        if not key in cloud_output:
            raise KeyNotInCloudOutput(
                f"Can not find Key {key} in Cloud Output for {resource_type}::{resource_hash} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        return cloud_output.get(key)


    def get_cloud_output_by_name(
        self,
        resource_state_uuid: str,
        component_name: str,
        resource_type: str,
        resource_name: str,
    ) -> Any:

        component = self.get_component(resource_state_uuid, component_name)
        resource = self.get_resource_by_name(
            resource_state_uuid, component_name, resource_type, resource_name
        )

        cloud_output_id = self._get_cloud_output_id(resource)

        if not cloud_output_id in component.cloud_output:

            raise CloudOutputDoesNotExist(
                f"Can not find Cloud Output for {resource_type}::{resource_name} in Component {component_name} in Resource State {resource_state_uuid}"
            )

        cloud_output = component.cloud_output.get(cloud_output_id)

        return cloud_output

    def create_differences(
        self,
        resource_state_uuid: str,
        new_components: List[ComponentModel],
        old_components: List[str],
    ) -> Tuple[
        Component_Difference, Resource_Reference_Difference, Resource_Difference
    ]:
        try:
            # Load the previous components
            previous_components: List[ComponentModel] = [
                self.get_component(resource_state_uuid, x) for x in old_components
            ]
           
        except Exception as e:
            raise e

        return _create_differences(new_components, previous_components)

    def _update_component(
        self,
        component: ComponentModel,
        diff: Resource_Difference,
        new_cloud_output: Dict = {},
        resolved_cloud_information: Dict={}
    ) -> ComponentModel:
        """
        Apply a resource difference over a component model and return the updated component model

        Arguments:
            component(ComponentModel): The component to update
            diff (Resource_Difference): The difference to apply
            new_cloud_output (Dict): The updated output if needed
            resolved_cloud_information (Dict): cloud output information used to deploy resource

        Returns:
            new_component(ComponentModel): The update component model

        """
        if diff.action_type == Resource_Change_Type.DELETE:
            
            component.resources = [
                x
                for x in component.resources
                if not (x.ruuid == diff.previous_resource.ruuid and x.name == diff.previous_resource.name)
            ] 
            
            # remove the previous resource's cloud output
            previous_resource_cloud_output_id = self._get_cloud_output_id(
                diff.previous_resource
            )
            
            if previous_resource_cloud_output_id in component.cloud_output:
                component.cloud_output.pop(previous_resource_cloud_output_id)


            if previous_resource_cloud_output_id  in component.previous_resolved_cloud_values:
                component.previous_resolved_cloud_values.pop(previous_resource_cloud_output_id)
            

        elif (
            diff.action_type == Resource_Change_Type.UPDATE_IDENTITY
            or diff.action_type == Resource_Change_Type.UPDATE_NAME
        ):
                   
            component.resources = [
                x
                for x in component.resources
                if not (x.ruuid == diff.previous_resource.ruuid and x.name == diff.previous_resource.name)
            ]  +  [diff.new_resource]

            # remove the previous resource's cloud output
            previous_resource_cloud_output_id = self._get_cloud_output_id(
                diff.previous_resource
            )

            if previous_resource_cloud_output_id in component.cloud_output:
                component.cloud_output.pop(previous_resource_cloud_output_id)

            if previous_resource_cloud_output_id  in component.previous_resolved_cloud_values:
                component.previous_resolved_cloud_values.pop(previous_resource_cloud_output_id)
            
            cloud_output_id = self._get_cloud_output_id(diff.new_resource)
            component.cloud_output[cloud_output_id] = new_cloud_output

            component.previous_resolved_cloud_values[cloud_output_id] = resolved_cloud_information
        

        elif diff.action_type == Resource_Change_Type.CREATE:
            component.resources.append(diff.new_resource)            

            cloud_output_id = self._get_cloud_output_id(diff.new_resource)

            component.cloud_output[cloud_output_id] = new_cloud_output
            component.previous_resolved_cloud_values[cloud_output_id] = resolved_cloud_information

        # recompute hash
        component.hash = _compute_component_hash(component)

        return component

    def _get_cloud_output_id(self, resource: ResourceModel) -> str:
        """
        Uniform way of generating cloud mapping id's

        Arguments:
            resource (ResourceModel): resource to get id of

        Returns:
            cloud_output_id (str): id for the resource
        """
        return f"{resource.ruuid};{resource.name}"

    def _get_resource_state_file_location(self, resource_state_uuid: str) -> FilePath:
        return self._central_state.resource_state_locations.get(resource_state_uuid)


# Helper functions
def _compute_component_hash(component: ComponentModel) -> str:
    """
    Uniform way of computing a component's identity hash

    Argument:
        component (ComponentModel): The component to compute the hash of

    Returns:
        hash (str): identity hash for the component
    """
    if component.resources:
        
        resources = [x for x in component.resources]
        resources.sort(key=lambda x: x.hash)

        resource_hashes = [x.hash for x in resources]
    else:
        resource_hashes = []

    if component.references:
        # TODO create hash of all the things
        references = [x for x in component.references]
        references.sort(key=lambda x: x.name)

        references_hashes = [x.name for x in references]
    else:
        references_hashes = []

    all_hashes = references_hashes + resource_hashes

    return cdev_hasher.hash_list(all_hashes)


def _create_resource_diffs(
    component_name: str,
    new_resources: List[ResourceModel],
    old_resource: List[ResourceModel],
) -> List[Resource_Difference]:

    if old_resource:
        # build map<hash,resource>
        old_hash_to_resource: Dict[str, ResourceModel] = {
            x.hash: x for x in old_resource
        }
        # build map<name,resource>
        old_name_to_resource: Dict[str, ResourceModel] = {
            x.name: x for x in old_resource
        }
    else:
        old_hash_to_resource = {}
        old_name_to_resource = {}


    rv = []
    for resource in new_resources:
        if (
            resource.hash in old_hash_to_resource
            and resource.name in old_name_to_resource
        ):

            # POP the seen previous resources as we go so only remaining resources will be deletes
            old_resource.remove(old_hash_to_resource.get(resource.hash))
            continue

        elif (
            resource.hash in old_hash_to_resource
            and not resource.name in old_name_to_resource
        ):

            rv.append(
                Resource_Difference(
                    **{
                        "action_type": Resource_Change_Type.UPDATE_NAME,
                        "component_name": component_name,
                        "previous_resource": old_hash_to_resource.get(resource.hash),
                        "new_resource": resource,
                    }
                )
            )
            # POP the seen previous resources as we go so only remaining resources will be deletes
            old_resource.remove(old_hash_to_resource.get(resource.hash))

        elif (
            not resource.hash in old_hash_to_resource
            and resource.name in old_name_to_resource
        ):

            rv.append(
                Resource_Difference(
                    **{
                        "action_type": Resource_Change_Type.UPDATE_IDENTITY,
                        "component_name": component_name,
                        "previous_resource": old_name_to_resource.get(resource.name),
                        "new_resource": resource,
                    }
                )
            )
            # POP the seen previous resources as we go so only remaining resources will be deletes
            old_resource.remove(old_name_to_resource.get(resource.name))

        elif (
            not resource.hash in old_hash_to_resource
            and not resource.name in old_name_to_resource
        ):
            
            rv.append(
                Resource_Difference(
                    **{
                        "action_type": Resource_Change_Type.CREATE,
                        "component_name": component_name,
                        "previous_resource": None,
                        "new_resource": resource,
                    }
                )
            )

    if old_resource:
        for resource in old_resource:
           
            rv.append(
                Resource_Difference(
                    **{
                        "action_type": Resource_Change_Type.DELETE,
                        "component_name": component_name,
                        "previous_resource": resource,
                        "new_resource": None,
                    }
                )
            )

    return rv


def _create_reference_diffs(
    new_references: List[ResourceReferenceModel],
    old_references: List[ResourceReferenceModel],
    originating_component_name: str, 
) -> List[Resource_Reference_Difference]:

    if old_references:
        # build map<name,resource>
        old_name_to_references: Dict[str, ResourceReferenceModel] = {
            f"{x.ruuid};;{x.name}": x for x in old_references
        }
    else:
        old_name_to_references = {}

    rv = []
    for reference in new_references:
        _id = f"{reference.ruuid};;{reference.name}"
        if _id in old_name_to_references:
            # POP the seen previous resources as we go so only remaining resources will be deletes
            old_references.remove(old_name_to_references.get(_id))

        else:
            rv.append(
                Resource_Reference_Difference(
                    Resource_Reference_Change_Type.CREATE, originating_component_name, reference
                )
            )

    for old_reference in old_references:
        
        rv.append(
            Resource_Reference_Difference(
                Resource_Reference_Change_Type.DELETE, originating_component_name, old_reference
            )
        )

    return rv


def _create_differences(
    new_components: List[ComponentModel], previous_components: List[ComponentModel]
) -> Tuple[
    List[Component_Difference],
    List[Resource_Difference],
    List[Resource_Reference_Difference]
]:

    component_diffs = []
    resource_diffs = []
    reference_diffs = []

    if previous_components:
        # build map<hash,resource>
        previous_hash_to_component = {x.hash: x for x in previous_components}
        # build map<name,resource>
        previous_name_to_component = {x.name: x for x in previous_components}
        previous_components_to_remove = [x for x in previous_components]
    else:
        previous_hash_to_component = {}
        previous_name_to_component = {}
        previous_components_to_remove = []

    if new_components:
        for component in new_components:

            if (
                not component.hash in previous_hash_to_component
                and not component.name in previous_name_to_component
            ):
                # Create component and all resources and all references
            
                component_diffs.append(
                    Component_Difference(
                        Component_Change_Type.CREATE, new_name=component.name
                    )
                )

                tmp_resource_diff = _create_resource_diffs(
                    component.name, component.resources, []
                )
                resource_diffs.extend(tmp_resource_diff)

                tmp_reference_diff = _create_reference_diffs(component.references, [], component.name)
                reference_diffs.extend(tmp_reference_diff)

            elif (
                component.hash in previous_hash_to_component
                and component.name in previous_name_to_component
            ):
                # Since the hash is the same we can infer all the resource hashes and reference hashes are the same
                # Even though the hash has remained the same we need to check for name changes in the resources
                previous_component = previous_name_to_component.get(component.name)

                # Should only output resource name changes
                tmp_resource_diff = _create_resource_diffs(
                    component.name, component.resources, previous_component.resources
                )

                if any(
                    [
                        not x.action_type == Resource_Change_Type.UPDATE_NAME
                        for x in tmp_resource_diff
                    ]
                ):
                    # if there is a resource change that is not an update name raise an exception
                    raise Exception

                resource_diffs.extend(tmp_resource_diff)

                # POP the seen previous component as we go so only remaining resources will be deletes
                previous_components_to_remove.remove(previous_component)

            elif (
                component.hash in previous_hash_to_component
                and not component.name in previous_name_to_component
            ):
                # hash of the component has stayed the same but the user has renamed the component name
                # Even though the hash has remained the same we need to check for name changes in the resources
                previous_component = previous_hash_to_component.get(component.hash)

                component_diffs.append(
                    Component_Difference(
                        Component_Change_Type.UPDATE_NAME,
                        previous_name=previous_component.name,
                        new_name=component.name,
                    )
                )

                # Should only output resource name changes
                tmp_resource_diff = _create_resource_diffs(
                    component.name, component.resources, previous_component.resources
                )

                if any(
                    [
                        not x.action_type == Resource_Change_Type.UPDATE_NAME
                        for x in tmp_resource_diff
                    ]
                ):
                    # if there is a resource change that is not an update name raise an exception
                    raise Exception

                resource_diffs.extend(tmp_resource_diff)

                # POP the seen previous component as we go so only remaining resources will be deletes
                previous_components_to_remove.remove(previous_component)

            elif (
                not component.hash in previous_hash_to_component
                and component.name in previous_name_to_component
            ):
                # hash of the component has changed but not the name
                # This means a resource or reference has updated its identity hash
                previous_component = previous_name_to_component.get(component.name)

                tmp_resource_diff = _create_resource_diffs(
                    component.name, component.resources, previous_component.resources
                )
                resource_diffs.extend(tmp_resource_diff)

                tmp_reference_diff = _create_reference_diffs(
                    component.references, previous_component.references, component.name
                )
                reference_diffs.extend(tmp_reference_diff)

                component_diffs.append(
                    Component_Difference(
                        Component_Change_Type.UPDATE_IDENTITY,
                        previous_name=previous_component.name,
                        new_name=component.name,
                    )
                )

                # POP the seen previous component as we go so only remaining resources will be deletes
                previous_components_to_remove.remove(previous_component)

    for removed_component in previous_components_to_remove:

        component_diffs.append(
            Component_Difference(
                Component_Change_Type.DELETE,
                previous_name=removed_component.name,
            )
        )

        tmp_resource_diff = _create_resource_diffs(
            removed_component.name, [], removed_component.resources
        )
        resource_diffs.extend(tmp_resource_diff)

        tmp_reference_diff = _create_reference_diffs([], removed_component.references, removed_component.name)
        reference_diffs.extend(tmp_reference_diff)

    return component_diffs, resource_diffs, reference_diffs



from ..utils.exceptions import Cdev_Error


class BackendError(Cdev_Error):
    pass


class ResourceStateAlreadyExists(BackendError):
    pass


class ResourceStateDoesNotExist(BackendError):
    pass


class ResourceStateNotEmpty(BackendError):
    pass


class ComponentAlreadyExists(BackendError):
    pass


class ComponentDoesNotExist(BackendError):
    pass


class ComponentNotEmpty(BackendError):
    pass


class ResourceChangeTransactionDoesNotExist(BackendError):
    pass


class ResourceDoesNotExist(BackendError):
    pass


class CloudOutputDoesNotExist(BackendError):
    pass


class KeyNotInCloudOutput(BackendError):
    pass


class ResourceReferenceError(BackendError):
    pass



import heapq
import json
import math
import os
from pydantic.types import DirectoryPath, FilePath
import shutil
from typing import List, Dict, Tuple, Set
from zipfile import ZipFile
from core.constructs.workspace import Workspace


from core.utils import paths as core_paths, hasher as cdev_hasher
from core.utils.logger import log
from core.default.resources.simple.xlambda import  DependencyLayer


from . import utils as fs_utils
from .utils import PackageTypes, ModulePackagingInfo, ExternalDependencyWriteInfo
from .external_dependencies_index import weighted_dependency_graph, compute_index



EXCLUDE_SUBDIRS = {"__pycache__"}



class LayerWriterCache:
    """
    Naive cache implentation for writing zip files for the lambda layers.
    """

    def __init__(self, cache_location: str) -> None:
        self._cache_location = cache_location
        if not os.path.isfile(self._cache_location):
            self._cache = {}

        else:
            with open(self._cache_location) as fh:
                self._cache = json.load(fh)

    def find_item(self, id: str):
        return self._cache.get(id)

    def add_item(self, id: str, item: json):
        self._cache[id] = item

        with open(self._cache_location , "w") as fh:
            json.dump(self._cache, fh, indent=4)


LAYER_CACHE = LayerWriterCache(os.path.join(os.getcwd(),".cache") )


def create_full_deployment_package(
    original_path: FilePath,
    needed_lines: List[int],
    function_name: str,
    base_archive_directory: DirectoryPath,
    pkgs: Dict[str, ModulePackagingInfo] = None,
    available_layers: int = 5,
):
    """
    Create all the needed deployment resources needed for a given serverless function. This includes parsing out the needed lines from
    the original function, packaging local dependencies, and packaging external dependencies (pip packages).

    Args:
        original_path (FilePath): The original file location that this function is from
        needed_lines (List[int]): The list of line numbers needed from the original function
        function_name (str): The name of the function
        base_archive_directory (DirectoryPath): Base path to write final archives
        pkgs (Dict[str, ModulePackagingInfo]): The list of package info for the function
        available_layers (int): The number of layers available to use

    Returns:
        handler_archive_location (FilePath): The location of the created archive for the handler
        handler_archive_hash (str): Identifying hash of the handler archive
        base_handler_path (str): The path to the file as a python package path. This will be the handler entry point when deployed.
        external_dependencies (LambdaLayerArtifact): List of layers that are needed for this function
    """
    global LAYER_CACHE

    LAYER_CACHE = LayerWriterCache(os.path.join(base_archive_directory, "cache"))

    # Clean the name of the function and create the final path of the parsed function
    cleaned_name = function_name.replace(" ", "_")   
    parsed_path = fs_utils.get_parsed_path(original_path, cleaned_name, base_archive_directory)
    log.debug("Parsed Path (%s) for %s/%s", parsed_path, original_path, cleaned_name)

    # Write the actual parsed function into an intermediate folder
    _make_intermediate_handler_file(original_path, needed_lines, parsed_path)

    # Keeps track of all files that need to be include in the handler archive
    handler_files = [parsed_path]

    # The handler can be in a subdirectory and since we preserve the relative project structure, we need to bring any __init__.py files
    # to make sure the handler is in a valid path
    extra_handler_path_files = _find_packaging_files_handler(original_path, base_archive_directory)

    handler_files.extend(extra_handler_path_files)

    # Start from the base archive folder location then replace '/' with '.' and final remove the '.py' from the end
    base_handler_path = os.path.relpath(
        parsed_path,
        start=base_archive_directory
    ).replace("/", ".")[:-3]

    # Zip archive will be at the same directory as the handler and replaced '.py' with '.zip
    parsed_filename = os.path.split(parsed_path)[1]
    zip_archive_location = os.path.join(
        os.path.dirname(parsed_path), parsed_filename[:-3] + ".zip"
    )

    dependencies_info: List[DependencyLayer] = []

    if pkgs:
        # Based on the directly references modules in the handler
        # Figure out the optimal modules to package to need the needs. 
        layer_dependencies, handler_dependencies = _create_package_dependencies_info(
            pkgs
        )

        if layer_dependencies:
            # Make the layer archive in the same folder as the handler

            LAYERS_DIR_NAME = "layers"

            archive_dir = os.path.join(
                base_archive_directory, LAYERS_DIR_NAME
            )

            if not os.path.isdir(archive_dir):
                os.mkdir(archive_dir)

            (
                single_dependency_layers,
                composite_layer,
            ) = _create_layers_from_referenced_modules(
                layer_dependencies, available_layers
            )
                    

            if single_dependency_layers:
                for single_layer in single_dependency_layers:
                    dependencies_info.append(
                        _make_single_dependency_archive(single_layer, archive_dir)
                    )

            if composite_layer:
                dependencies_info.append(
                    _make_composite_dependency_archive(composite_layer, archive_dir)
                )

        if handler_dependencies:
            # Copy the local dependencies files into the intermediate folder to make packaging easier
            # All the local copied files are added to the set of files needed to be include in the .zip file uploaded as the handler
            # Add their copied locations into the handler_files var so that they are written to the final handler archive
            local_dependencies_intermediate_locations = _copy_local_dependencies(
                handler_dependencies,
                base_archive_directory
            )
            handler_files.extend(local_dependencies_intermediate_locations)


    # Create the actual handler archive by zipping the needed files
    archive_hash = _make_intermediate_handler_zip(zip_archive_location, handler_files, base_archive_directory)

    return (zip_archive_location, archive_hash, base_handler_path, dependencies_info)


def _create_package_dependencies_info(
    directly_referenced_pkgs: Dict[str, ModulePackagingInfo]
) -> Tuple[List[ModulePackagingInfo], List[str]]:
    """
    Take a dictionary of the directly referenced packages (str [pkg_name] -> ModulePackagingInfo) that are used by a handler and return the
    necessary information for creating the deployment packages.

    Args:
        pkgs (Dict[str, ModulePackagingInfo]): Directly referenced packages used by the handler

    Returns:
        layer_dependencies (List[ModulePackagingInfo]): The packages to be added to the layers
        handler_dependencies (List[str]): List of files to include with the handler
    
    Packages can be broken down into two categories: handler and layer.

    
    Handler packages are located within the `Workspace`, and therefore, should be packaged into the handler archive so that they
    remain in the correct relative location to the handler function.
        src:\n
        |_ views\n
        |___ handlers.py\n
        |_ models\n
        |___ model.py\n

    If a function in the handlers.py file references models.py as a relative package (from .. import models) it is important to keep 
    the relative file structure the same.

    Layer packages are packages that are found on the PYTHONPATH and installed with a package manager like PIP. These
    should be packages as layers to keep the handler archive size small. Read the Cdev dependency deep dives for an in depth breakdown of
    how the layers are being packaged.

    Note that the input are the directly referenced packages used by the handler, so we must look at the 'flat' attribute to find all the
    needed dependencies.
    """

    handler_dependencies = set()
    directly_referenced_module_write_info: Set[ModulePackagingInfo] = set()

    for _, pkg in directly_referenced_pkgs.items():

        if pkg.type == PackageTypes.PIP:
            directly_referenced_module_write_info.add(pkg)

        elif pkg.type == PackageTypes.LOCALPACKAGE:
            full_path  = core_paths.get_full_path_from_workspace_base(pkg.fp)
            if core_paths.is_in_workspace(full_path):
                if os.path.isdir(pkg.fp):
                    # If the fp is a dir that means we need to include the entire directory
                    for dir, _, files in os.walk(pkg.fp):
                        if dir.split("/")[-1] in EXCLUDE_SUBDIRS:
                            # If the directory is in the excluded subdirs then we can just continue
                            # i.e __pycache__
                            continue

                        handler_dependencies = handler_dependencies.union(
                            set([os.path.join(pkg.fp, dir, x) for x in files])
                        )
                else:
                    handler_dependencies.add(pkg.fp)
            else:
                # Can not link to a locally managed module that is not within the cdev project structure.
                print(f"Linking to local module outside of Workspace {pkg.fp}")
                raise Exception

            if pkg.flat:
                # If the local module has dependencies, we need to recursively search them for import statements
                directly_referenced_module_write_info.update(
                    _recursively_find_directly_referenced_modules_in_local_module(pkg)
                )

                for dependency in pkg.flat:
                    if dependency.type == PackageTypes.LOCALPACKAGE:
                        handler_dependencies.add(dependency.fp)

    return (list(directly_referenced_module_write_info), list(handler_dependencies))


def _create_layers_from_referenced_modules(
    directly_referenced_modules: List[ModulePackagingInfo], available_layers: int = 2
) -> Tuple[List[ModulePackagingInfo], List[ModulePackagingInfo]]:
    """
    Run the optimization calculations for creating single dependencies and composite dependencies for the external modules used. Returns the list of modules that should be
    in single dependencies and in the composite layer.

    Args:
        directly_referenced_modules (List[ModulePackagingInfo]): The modules that are directly reference
        available_layers (int): Amount of layers that are available to be used to package the dependencies

    Returns:
        Single Dependency Layers (List[ModulePackagingInfo]): The list of dependencies that should be individual layers
        Composite Dependency Layers (List[ModulePackagaingInfo]): The list of dependencies that go into the composite layer
    """

    # Now that we have all the directly referenced import statements that are for packaged modules, we compute a weighted DAG to create
    # an optimized deployment platform
    # For a more detailed look at this optimization process read the dependency deep dive on the cdev website.
    graph = weighted_dependency_graph(directly_referenced_modules)


    directly_referenced_modules_removed = set(directly_referenced_modules).difference(
        graph.true_top_level_modules
    )


    # Optimization algorithm:
    # Since there is only one available layer, everything must go into the composite layer
    if available_layers == 1:
        return [], graph.true_top_level_modules

    # if the true top level modules is less than (or equal to) the available layers then use a layer for each top level module
    # so that the top level module layer can be reused within the project
    elif len(graph.true_top_level_modules) <= available_layers:
        return graph.true_top_level_modules, []

    # If there are not enough layers available for all the top level modules, we need to create a composite layer that contains more
    # than one top level module. We will create 1 composite layer and N amount of single layers such that the single layers remove the
    # most amount of code that needs to be uploaded in the composite layer.
    else:
        max_index_level = 2
        # We want the index to be the lower of the two:
        # The user defined max index size (so they can cap computer and memory time of the index)
        # The available layers that can be used for single dependencies
        index_depth = min(max_index_level, available_layers)

        
        index = compute_index(graph, index_depth)
        for pair, value in index.items():
            print(f"{pair} -> {math.floor(value/1024)} KB")

        index_heap = [(-value, key) for key, value in index.items()]
        heapq.heapify(index_heap)

        single_dependency_layer_ids = heapq.heappop(index_heap)[1]

        single_dependency_layers: List[ModulePackagingInfo] = []
        composite_layer: List[ModulePackagingInfo] = []

        for dependency in graph.true_top_level_modules:
            if dependency.get_id_str() in single_dependency_layer_ids:
                single_dependency_layers.append(dependency)
            else:
                composite_layer.append(dependency)

        return single_dependency_layers, composite_layer


def _recursively_find_directly_referenced_modules_in_local_module(
    local_module: ModulePackagingInfo,
) -> Set[ModulePackagingInfo]:
    rv = set()
    if local_module.tree:
        for child in local_module.tree:
            if child.type == PackageTypes.PIP:
                rv.add(child)

            elif child.type == PackageTypes.LOCALPACKAGE:
                rv.update(
                    _recursively_find_directly_referenced_modules_in_local_module(child)
                )

    return rv


def _make_intermediate_handler_file(
    original_path: FilePath, needed_lines: List[int], parsed_path: str
):
    """
    Make the actual file that will be deployed onto a serverless platform by parsing out the needed lines from the original file
    and writing them to an intermediate file

    Args:
        original_path (FilePath): The original file location that this function is from
        needed_lines (List[int]): The list of line numbers needed from the original function
        parsed_path (str): The final location of the parsed file
    """
    if not os.path.isfile(original_path):
        raise Exception

    file_list = fs_utils.get_file_as_list(original_path)

    actual_lines = fs_utils.get_lines_from_file_list(file_list, needed_lines)

    cleaned_actual_lines = _clean_lines(actual_lines)

    _write_intermediate_function(parsed_path, cleaned_actual_lines)


def _find_packaging_files_handler(original_path: FilePath, base_intermediate_path: DirectoryPath) -> List[FilePath]:
    """
    This adds additional files like __init__.py that are needed to make the file work. Since the handler can be in submodules, 
    it is important to find all the nested __init__.py files for making sure the handler is accessible.

    Args:
        original_path (FilePath): The original path of the file

    Returns:
        needed_files (List[Filepath]): The list of files that need to be included with the handler
    """
    path_from_project_dir = os.path.dirname(
        core_paths.get_relative_to_workspace_path(original_path)
    ).split("/")

    tmp = ""
    actual_base_path = Workspace.instance().settings.BASE_PATH
    rv = []

    # Traverse the directory path looking for __init__.py files. If an __init__.py is found then copy it over
    # else just make an empty one   
    while path_from_project_dir:

        tmp = os.path.join(
            tmp, path_from_project_dir.pop(0)
        )

        file_loc = os.path.join(actual_base_path, tmp, "__init__.py")

        intermediate_file_location = os.path.join(base_intermediate_path, tmp, "__init__.py")

        if os.path.isfile(file_loc):
            shutil.copyfile(file_loc, intermediate_file_location)
        else:
            with open(intermediate_file_location, "a"):
                os.utime(intermediate_file_location)

        rv.append(intermediate_file_location)

    return rv


def _clean_lines(lines: List[str]) -> List[str]:
    """
    Parsed functions can have empty lines or comments as the last lines of the file, so we are going to start from the end of the file
    and remove those lines. This helps keep the hashes of the handler consistent even if there was changes below the handler that
    are picked up because they are comments.


    Args:
        lines (List[str]): original lines to be added for the handler

    Returns:
        clean_lines (List[str]): Lines with endings trimmed of whitespace and comments

    """

    # final line should be an offset from the end of the list the represents the final real line of python code
    final_line_no = -1

    for i in range(len(lines) - 1):
        tmp_line = lines[-(i + 1)]

        # if the line is blank it is not a valid line
        if not tmp_line:
            continue

        # if the line is a comment (starts with '#') or is just whitespace
        if not (tmp_line[0] == "#" or tmp_line.isspace()):
            final_line_no = i
            break

    if final_line_no == -1 or final_line_no == 0:
        rv = lines
    else:
        rv = lines[:-final_line_no]

    return rv


def _copy_local_dependencies(dependencies: List[FilePath], base_final_dir: DirectoryPath) -> List[FilePath]:
    """
    Copy the local dependency files from their original location into the intermediate folder with the actual handler.
    This step makes the archiving step simplier since all the need files are in the intermediate folder.

    Args:
        file_locations (List[str]): list of file locations that need to be copied
        base_final_dir (DirectoryPath): base path for copied files

    Returns:
        copied_file_locations (List[str]): list of locations of the copied files
    """
    rv = []


    for dependency in dependencies:
        intermediate_location = os.path.join(
            base_final_dir,
            core_paths.get_relative_to_workspace_path(dependency)
        )

        relative_to_intermediate = core_paths.get_relative_to_intermediate_path(
            intermediate_location
        ).split("/")[:-1]

        core_paths.create_path(base_final_dir, relative_to_intermediate)

        if os.path.isdir(dependency):
            if os.path.isdir(intermediate_location):
                shutil.rmtree(intermediate_location)

            shutil.copytree(
                dependency,
                intermediate_location,
                ignore=shutil.ignore_patterns("*.pyc"),
            )

            for dir_name, _, files in os.walk(intermediate_location):
                for file in files:
                    rv.append(os.path.join(dir_name, file))

        elif os.path.isfile(dependency):
            shutil.copyfile(dependency, intermediate_location)
            rv.append(intermediate_location)
        else:
            raise Exception

    return rv


def _write_intermediate_function(path: FilePath, lines: List[str]):
    """
    Write a set of lines for the intermediate handler into the actual file on the system

    Args:
        path (FilePath): Intermediate file path
        lines (List[str]): List of the lines to write
    """

    with open(path, "w") as fh:
        for line in lines:
            fh.write(line)
            fh.write("\n")


def _make_intermediate_handler_zip(
    zip_archive_location: str, paths: List[FilePath], base_path: DirectoryPath
) -> str:
    """
    Make the archive for the handler deployment.

    Args:
        zip_archive_location (str): The file path for the archive. Might not be created yet
        paths (List[FilePath]): The list of files that should be written to the archive

    Returns:
        archive_hash (str): An identifying hash for the zip
    """
    hashes = []
    paths.sort()
    with ZipFile(zip_archive_location, "w") as zipfile:
        for path in paths:
            filename = os.path.relpath(path, base_path)
            zipfile.write(path, filename)
            hashes.append(cdev_hasher.hash_file(path))

    return cdev_hasher.hash_list(hashes)


def _make_single_dependency_archive(
    module_info: ModulePackagingInfo, archive_dir: DirectoryPath
) -> DependencyLayer:
    needed_info: List[ExternalDependencyWriteInfo] = []
    

    needed_info.append(
        ExternalDependencyWriteInfo(
            **{"location": module_info.fp, "id": module_info.get_id_str()}
        )
    )

    for child in module_info.flat:
        if not child.type == PackageTypes.PIP:
            continue

        needed_info.append(
            ExternalDependencyWriteInfo(
                **{"location": child.fp, "id": child.get_id_str()}
            )
        )

    # Create a hash of the ids of the packages to see if there is an already available archive to use
    ids = [x.id for x in needed_info]
    ids.sort()

    _id_hashes = cdev_hasher.hash_list(ids)
    cache_item = LAYER_CACHE.find_item(_id_hashes)

    layer_name = module_info.get_id_str()

    if cache_item:
        return DependencyLayer(
            cdev_name=layer_name, artifact_path=cache_item[0], artifact_hash=cache_item[1]
        )

    archive_path, archive_hash = _make_layers_zips(archive_dir, layer_name, needed_info)

    LAYER_CACHE.add_item(_id_hashes, (archive_path, archive_hash))

    return DependencyLayer(cdev_name=layer_name, artifact_path=archive_path, artifact_hash=archive_hash)


def _make_composite_dependency_archive(
    module_info: List[ModulePackagingInfo], archive_dir: DirectoryPath
) -> DependencyLayer:
    needed_info: Set[ExternalDependencyWriteInfo] = set()

    print(f"Making Composite layer")
    for module in module_info:
        needed_info.add(
            ExternalDependencyWriteInfo(
                **{"location": module.fp, "id": module.get_id_str()}
            )
        )

        for child in module.flat:
            print(f"Adding  {child}")
            if not child.type == PackageTypes.PIP:
                continue

            needed_info.add(
                ExternalDependencyWriteInfo(
                    **{"location": child.fp, "id": child.get_id_str()}
                )
            )

    # Create a hash of the ids of the packages to see if there is an already available archive to use
    ids = [x.id for x in needed_info]
    ids.sort()

    _id_hashes = cdev_hasher.hash_list(ids)
    cache_item = LAYER_CACHE.find_item(_id_hashes)
    if cache_item:
        return None

    layer_name = f"composite_{_id_hashes}"

    archive_path, archive_hash = _make_layers_zips(
        archive_dir, layer_name, list(needed_info)
    )

    # convert to json string then back to python object because it has a Filepath type in it and that is always handled weird.
    LAYER_CACHE.add_item(_id_hashes, (archive_path, archive_hash))

    rv = DependencyLayer(cdev_name=layer_name, artifact_path=archive_path)

    return rv


def _make_layers_zips(
    zip_archive_location_directory: DirectoryPath,
    layer_name: str,
    needed_info: List[ExternalDependencyWriteInfo],
) -> Tuple[FilePath, str]:
    """Create the zip archive that will be deployed with the handler function.
    
    All modules provide are written to a single archive such that the module is in '/python/<module_name>'.

    Args:
        zip_archive_location_directory (DirectoryPath): The directory that the archive will be created in.
        layer_name (str): base name for the archive
        needed_info (List[ExternalDependencyWriteInfo]): The information about what modules to add to the archive

    Returns:
        archive_path (str): The location of the created archive relative to the Cdev project location
        hash (str): A identifying hash of the archive
    """

    _file_hashes = []
    zip_archive_full_path = os.path.join(
        zip_archive_location_directory, layer_name + ".zip"
    )
    seen_pkgs = set()

    if os.path.isfile(zip_archive_full_path):
        os.remove(zip_archive_full_path)

    for info in needed_info:
        if info.id in seen_pkgs:
            continue

        seen_pkgs.add(info.id)

        with ZipFile(zip_archive_full_path, "a") as zipfile:
            if os.path.isfile(info.location):
                # this is a single python file not a folder (ex: six.py)
                file_name = os.path.split(info.location)[1]
                # since this is a module that is just a single file plop in /python/<filename> and it will be on the pythonpath

                zipfile.write(info.location, os.path.join("python", file_name))
                _file_hashes.append(cdev_hasher.hash_file(info.location))

            else:
                pkg_name = os.path.split(info.location)[1]

                for dirname, subdirs, files in os.walk(info.location):
                    if dirname.split("/")[-1] in EXCLUDE_SUBDIRS:
                        continue

                    zip_dir_name = os.path.normpath(
                        os.path.join(
                            "python", pkg_name, os.path.relpath(dirname, info.location)
                        )
                    )

                    for filename in files:
                        original_path = os.path.join(dirname, filename)
                        zipfile.write(
                            original_path, os.path.join(zip_dir_name, filename)
                        )
                        _file_hashes.append(cdev_hasher.hash_file(original_path))

                pkg_dir = os.path.dirname(info.location)
                for obj in os.listdir(pkg_dir):
                    # Search in the general packaging directory for any other directory with the package name
                    # for example look for numpy.lib when including numpy
                    if os.path.join(pkg_dir, obj) == info.location:
                        continue

                    if (
                        os.path.isdir(os.path.join(pkg_dir, obj))
                        and obj.split(".")[0] == os.path.split(info.location)[1]
                    ):
                        for dirname, subdirs, files in os.walk(
                            os.path.join(pkg_dir, obj)
                        ):
                            if dirname.split("/")[-1] in EXCLUDE_SUBDIRS:
                                continue

                            zip_dir_name = os.path.normpath(
                                os.path.join(
                                    "python",
                                    obj,
                                    os.path.relpath(
                                        dirname, os.path.join(pkg_dir, obj)
                                    ),
                                )
                            )

                            for filename in files:
                                original_path = os.path.join(dirname, filename)
                                zipfile.write(
                                    original_path, os.path.join(zip_dir_name, filename)
                                )
                                _file_hashes.append(
                                    cdev_hasher.hash_file(original_path)
                                )

    full_archive_hash = cdev_hasher.hash_list(_file_hashes)

    return (
        zip_archive_full_path,
        full_archive_hash,
    )



from core.constructs.commands import BaseCommandContainer


class SimpleCommandContainer(BaseCommandContainer):
    help = """
These are the commands that are available for use on simple buckets.

cp: Copy files between your local workspace and a deployed bucket
sync: Sync the contents of a given folder to a deployed bucket

"""



"""Set of constructs for making Serverless Functions

"""

import importlib
import inspect
import os
from pydantic import FilePath
from typing import Any, Callable, Dict, FrozenSet, List, Optional, Union

from core.constructs.cloud_output import Cloud_Output, Cloud_Output_Dynamic, Cloud_Output_Str
from core.constructs.resource import Resource, ResourceModel, update_hash, ResourceOutputs, PermissionsGrantableMixin
from core.constructs.models import frozendict, ImmutableModel
from core.constructs.types import cdev_str_model, cdev_str

from core.utils import hasher
from core.utils.platforms import lambda_python_environment, get_current_closest_platform

from core.default.resources.simple.iam import Permission, PermissionArn, permission_arn_model, permission_model
from core.default.resources.simple.events import Event, event_model


LAMBDA_LAYER_RUUID = "cdev::simple::lambda_layer"
RUUID = "cdev::simple::function"

################
##### Dependencies 
################
class DeployedLayer():
    """Construct that represents a Layer already deployed on the Cloud"""
    def __init__(self, arn: str) -> None:
        self.arn = arn


class dependency_layer_model(ResourceModel):
    """Model that represents a local folder that will be deployed on the cloud as a Layer"""
    artifact_path: str


class DependencyLayer(Resource):
    """Construct that represents a local folder that will be deployed on the cloud as a Layer"""
    def __init__(self, cdev_name: str, artifact_path: FilePath, artifact_hash: str) -> None:
        super().__init__(cdev_name, LAMBDA_LAYER_RUUID)
        self.artifact_path = artifact_path
        self.artifact_hash = artifact_hash
        self._hash = artifact_hash
        self.output = DependencyLayerOutput(cdev_name)

    def render(self) -> dependency_layer_model:
        return dependency_layer_model(
            ruuid=LAMBDA_LAYER_RUUID,
            name=self.name, 
            hash=self.artifact_hash,
            artifact_path=self.artifact_path
        )

################
##### Output
################


class DependencyLayerOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a Layer has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, LAMBDA_LAYER_RUUID)


    @property
    def layer_arn(self) -> Cloud_Output_Str:
        """The id of the created layer"""
        return Cloud_Output_Str(
            name=self._name,
            ruuid=LAMBDA_LAYER_RUUID,
            key='arn',
            type=self.OUTPUT_TYPE
        )

    @layer_arn.setter
    def layer_arn(self, value: Any):
        raise Exception


class FunctionOutput(ResourceOutputs):
    """Container object for the returned values from the cloud after a Serverless Function has been deployed."""
    def __init__(self, name: str) -> None:
        super().__init__(name, RUUID)


################
##### Function
################
class simple_function_configuration_model(ImmutableModel):
    """Model representing the configuration of a Serverless Function"""
    handler: str
    description: Optional[cdev_str_model]
    environment_variables: frozendict

    class Config:
        use_enum_values = True
        # Beta Feature but should be fine since this is simple data 
        frozen = True


class SimpleFunctionConfiguration():
    """Container for the configuration settings of Serverless Function"""
    def __init__(self, handler: cdev_str, description: cdev_str = "", environment_variables: Dict[str, cdev_str] = {} ) -> None:
        """
        Args:
            handler (cdev_str): The python module path of the handler function that is triggered by the Cloud Platform
            description (cdev_str, optional): A description of the Function. Defaults to "".
            environment_variables (Dict[str, cdev_str], optional): A dict of overriding variabled for the Environment. Defaults to {}.
        """
        self.handler = handler
        self.description = description
        self.environment_variables = environment_variables


    def render(self) -> simple_function_configuration_model:
        return simple_function_configuration_model(
            handler=self.handler.render() if isinstance(self.handler, Cloud_Output_Dynamic) else self.handler,
            description=self.description.render() if isinstance(self.description, Cloud_Output_Dynamic) else self.description,
            environment_variables = frozendict(
                {k:frozendict(v.render().dict()) if isinstance(v, Cloud_Output) else v for k,v in self.environment_variables.items()}
            ) if self.environment_variables else frozendict({})
        )

    def hash(self) -> str:
        env_hashable = {k:(v if not isinstance(v, Cloud_Output) else v.hash()) for k,v in self.environment_variables.items()} 


        return hasher.hash_list([
            self.handler,
            self.description,
            hasher.hash_list(sorted(env_hashable.items()))
        ])



class simple_function_model(ResourceModel):
    """Model representing a Serverless Function

    Args:
        Filepath ([type]): [description]
    """
    filepath: str  # Don't use FilePath because this will be a relative path and might not always point correctly to a file in all contexts
    configuration: simple_function_configuration_model
    events: FrozenSet[event_model]
    permissions: FrozenSet[Union[permission_model, permission_arn_model]]
    external_dependencies: FrozenSet[Any]
    src_code_hash: str
    platform: lambda_python_environment



class SimpleFunction(PermissionsGrantableMixin, Resource):
    """Construct to represent a Serverless Function. It is recommend to generate this resource using the `simple_function_annotation`"""
    @update_hash
    def __init__(
        self,
        cdev_name: str,
        filepath: str,
        configuration: SimpleFunctionConfiguration,
        events: List[Event] = [],
        platform: lambda_python_environment = None,
        function_permissions: List[Union[Permission, PermissionArn]] = [],
        external_dependencies: List[Union[DeployedLayer, DependencyLayer]]=[],
        src_code_hash: str = None,
        nonce: str = "",
    ) -> None:
        """

        Args:
            cdev_name (str): Cdev Name for the function
            filepath (str): Path the final artifact to upload.
            configuration (SimpleFunctionConfiguration):  Configuration options for the function.
            events (List[Event], optional):  List of event triggers for the function.. Defaults to [].
            platform (lambda_python_environment, optional): Option to override the deployment platform in the Cloud.. Defaults to None.
            function_permissions (List[Union[Permission, PermissionArn]], optional): List of Permissions to grant to the Function.. Defaults to [].
            external_dependencies (List[Union[DeployedLayer, DependencyLayer]], optional): Dependencies to link to in the Cloud. Defaults to [].
            src_code_hash (str, optional): identifying hash of the source code. Defaults to None.
            nonce (str, optional): Nonce to make the resource hash unique if there are conflicting resources with same configuration.
        """
        super().__init__(cdev_name, RUUID, nonce)

        self._filepath = filepath
        self._events = events
        self._configuration = configuration
        self._external_dependencies = external_dependencies
        self._granted_permissions: List[Union[Permission, PermissionArn]] = function_permissions

        self.src_code_hash = src_code_hash if src_code_hash else hasher.hash_file(filepath)

        self._platform = platform if platform else get_current_closest_platform()
        

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    @update_hash
    def filepath(self, value: str):
        self._filepath = value

    @property
    def events(self):
        return self._events

    @events.setter
    @update_hash
    def events(self, value: List[Event]):
        self._events = value

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    @update_hash
    def configuration(self, value: SimpleFunctionConfiguration):
        self._configuration = value

    @property
    def platform(self):
        return self._platform

    @platform.setter
    @update_hash
    def platform(self, value: lambda_python_environment):
        self._platform = value

    @property
    def external_dependencies(self):
        return self._external_dependencies

    @external_dependencies.setter
    @update_hash
    def external_dependencies(self, value: List[Union[DeployedLayer, DependencyLayer]]):
        self._external_dependencies = value

    def compute_hash(self):
        self._permissions_hash = hasher.hash_list([x.hash() for x in self._granted_permissions])
        self._config_hash = self._configuration.hash()
        self._events_hash = hasher.hash_list([x.hash() for x in self.events])

        self._hash = hasher.hash_list(
            [
                self.src_code_hash,
                self._config_hash,
                self._events_hash,
                self._permissions_hash,
                self._nonce,
                self._platform
            ]
        )

    def render(self) -> simple_function_model:

        premissions = [x.render() for x in self.granted_permissions]

        dependencies = [x.output.cloud_id.render() if isinstance(x, DependencyLayer) else x.arn for x in self.external_dependencies]
        
        return simple_function_model(
            name=self.name,
            ruuid=self.ruuid,
            hash=self.hash,
            configuration=self.configuration.render(),
            filepath=self.filepath,
            events=frozenset([x.render() for x in self.events]),
            permissions=frozenset(premissions),
            external_dependencies=frozenset(dependencies),
            src_code_hash=self.src_code_hash,
            platform=self.platform
        )


def simple_function_annotation(
    name: str,
    events: List[Event] = [],
    environment={},
    permissions: List[Union[Permission, PermissionArn]] = [],
    override_platform: lambda_python_environment = None,
    includes: List[str] = [],
    nonce: str=""
) -> Callable[[Callable], SimpleFunction]:
    """This annotation is used to designate that a function should be deployed as a Serverless function.

    Args:
        name (str): Cdev Name for the function
        events (List[Event], optional): List of event triggers for the function. Defaults to [].
        environment (dict, optional): Dictionary to apply to the Environment Variables of the deployed function. Defaults to {}.
        permissions (List[Union[Permission, PermissionArn]], optional): List of Permissions to grant to the Function. Defaults to [].
        override_platform (lambda_python_environment, optional): Option to override the deployment platform in the Cloud. Defaults to None.
        includes (List[str], optional): Set of identifiers to extra global statements to include in parsed artifacts. Defaults to [].
        nonce (str, optional): Nonce to make the resource hash unique if there are conflicting resources with same configuration.

    Returns:
        Callable[[Callable], SimpleFunction]: wrapper that returns the `SimpleFunction`
    """

    def create_function(func: Callable) -> SimpleFunction:

        if inspect.isfunction(func):
            for item in inspect.getmembers(func):

                if item[0] == "__name__":
                    handler_name = item[1]
                elif item[0] == "__doc__":
                    description = item[1] if item[1] else ""
                elif item[0] == "__module__":
                    mod_name = item[1]


        final_config = SimpleFunctionConfiguration(
            handler=handler_name,
            description=description,
            environment_variables=environment
        )

        mod = importlib.import_module(mod_name)

        full_filepath = os.path.abspath(mod.__file__)

        return SimpleFunction(
            cdev_name=name,
            filepath=full_filepath,
            events=events,
            configuration=final_config,
            platform=override_platform,
            function_permissions=permissions,
            nonce=nonce
        )

    return create_function



import os
from re import S
from typing import Dict, List, Tuple, Set
from pydantic import BaseModel
from .utils import ExternalDependencyWriteInfo, ModulePackagingInfo, PackageTypes
from rich import print
from itertools import combinations
import math


class ExternalDependencyIndex(BaseModel):
    id: str
    table: Dict[str, int]
    level: int


class weighted_dependency_node:
    def __init__(self, id: str, individual_weight: int) -> None:
        self.id = id
        self.individual_weight = individual_weight
        self.total_weight = 0

        self._parents = set()
        self._children = set()

        self._all_children_flat = set()

    def set_total_weight(self, weight: int):
        self.total_weight = weight

    def add_parent(self, parent: "weighted_dependency_node"):
        self._parents.add(parent)

    def get_parents(self) -> Set["weighted_dependency_node"]:
        return self._parents

    def remove_parent(self, parent: "weighted_dependency_node"):
        if parent in self._parents:
            self._parents.remove(parent)

        else:
            raise Exception

    def add_child(self, child: "weighted_dependency_node"):
        self._children.add(child)
        child.add_parent(self)

        self.add_to_all_children(child.get_all_children())

    def get_children(self) -> Set["weighted_dependency_node"]:
        return self._children

    def get_all_children(self) -> Set["weighted_dependency_node"]:
        return self._all_children_flat

    def add_to_all_children(self, all_children: Set["weighted_dependency_node"]):
        self._all_children_flat.update(all_children)

    def __hash__(self) -> int:
        return int.from_bytes(self.id.encode(), "big")


_depth_to_color = {0: "white", 1: "blue", 2: "yellow", 3: "magenta", 4: "red"}


class weighted_dependency_graph:
    def __init__(self, top_level_modules: List[ModulePackagingInfo]) -> None:
        self._top_level_nodes: List[weighted_dependency_node] = []
        self._id_to_node: Dict[str, weighted_dependency_node] = {}
        self.HEAD = weighted_dependency_node("HEAD", -1)

        (
            self.true_top_level_modules,
            self.referenced_sub_modules,
        ) = self._create_pure_top_level_modules(top_level_modules)

        for top_level_module in self.true_top_level_modules:
            new_top_level_node = self._recursive_add_to_graph(top_level_module)

            if new_top_level_node:
                self._top_level_nodes.append(new_top_level_node)
                self.HEAD.add_child(new_top_level_node)

        for _, node in self._id_to_node.items():
            sub_graph_weight, _ = self._recursive_compute_total_subgraph_weight(
                node, set()
            )
            node.set_total_weight(sub_graph_weight)

    def _create_pure_top_level_modules(
        self, top_level_modules: List[ModulePackagingInfo]
    ) -> Tuple[List[ModulePackagingInfo], List[ModulePackagingInfo]]:
        all_child_modules = set()

        [
            *map(
                lambda x: all_child_modules.update({t.get_id_str() for t in x.flat})
                if x.flat
                else {},
                top_level_modules,
            )
        ]

        true_top_level_modules = []
        referenced_sub_modules = []

        for module_info in top_level_modules:
            if module_info.get_id_str() in all_child_modules:
                referenced_sub_modules.append(module_info)
            else:
                true_top_level_modules.append(module_info)

        return true_top_level_modules, referenced_sub_modules

    def _recursive_add_to_graph(
        self, node_info: ModulePackagingInfo
    ) -> weighted_dependency_node:
        if not node_info.type == PackageTypes.PIP:
            return None

        if not node_info.get_id_str() in self._id_to_node:

            module_individual_weight = get_module_size(node_info.fp)
            new_node = weighted_dependency_node(
                node_info.get_id_str(), module_individual_weight
            )

            if node_info.tree:
                for child_info in node_info.tree:
                    child_node = self._recursive_add_to_graph(child_info)

                    if child_node:
                        new_node.add_child(child_node)

            self._id_to_node[node_info.get_id_str()] = new_node

        else:
            new_node = self._id_to_node.get(node_info.get_id_str())

        return new_node

    def _recursive_compute_total_subgraph_weight(
        self, node: weighted_dependency_node, already_seen: Set[str]
    ) -> Tuple[int, Set[str]]:
        if node.id in already_seen:
            return 0, already_seen

        total_weight = node.individual_weight
        if node.get_children():
            for child in node.get_children():
                (
                    child_weight,
                    already_seen,
                ) = self._recursive_compute_total_subgraph_weight(child, already_seen)
                already_seen.add(child.id)
                total_weight += child_weight

        return total_weight, already_seen

    def print_graph(self):
        total_graph_weight = math.floor(self.get_total_weight() / 1024)
        max_graph_weight = math.floor((250 * 1024))
        percentage_used = math.floor((total_graph_weight / max_graph_weight) * 100)
        print(f"Total dependency size {total_graph_weight} kb ")
        print(f"    {percentage_used}% of max weight: {max_graph_weight} kb ")
        print(self._top_level_nodes)
        for node in self._top_level_nodes:
            total_weight_kb = math.floor(node.total_weight / 1024)
            individual_weight_kb = math.floor(node.individual_weight / 1024)

            print(f"{node.id}  ({total_weight_kb} kb) ({individual_weight_kb} kb)")
            self._recursive_print_graph(node, 0)

    def _recursive_print_graph(self, node: weighted_dependency_node, depth: int):
        for child in node.get_children():

            total_weight_kb = math.floor(child.total_weight / 1024)
            individual_weight_kb = math.floor(child.individual_weight / 1024)

            base_str = f"|[{_depth_to_color.get(depth%5)}]{'-' * depth } {child.id}[/{_depth_to_color.get(depth%5)}] ({total_weight_kb} kb) ({individual_weight_kb} kb) ({len(child.get_parents())})"
            print(base_str)

            self._recursive_print_graph(child, depth + 1)

    def get_total_weight(self) -> int:
        _already_seen_module = set()
        weight = 0

        for node in self._top_level_nodes:
            (
                subtree_unique_weight,
                _already_seen_module,
            ) = self._recursive_compute_weight(node, _already_seen_module)

            weight += subtree_unique_weight

        return weight

    def _recursive_compute_weight(
        self,
        node: weighted_dependency_node,
        already_seen: Set[weighted_dependency_node],
    ) -> Tuple[int, Set[weighted_dependency_node]]:
        weight = 0
        already_seen_module = already_seen
        if node in already_seen:
            # If we have already accounted for this nodes weight then continue
            return (0, already_seen)

        if node.get_children():
            for child in node.get_children():
                (
                    subtree_unique_weight,
                    already_seen_module,
                ) = self._recursive_compute_weight(child, already_seen_module)

                weight += subtree_unique_weight

        weight += node.individual_weight
        already_seen.update([node])

        return weight, already_seen


EXCLUDE_SUBDIRS = {"__pycache__"}


def get_module_size(starting_path: str) -> int:

    if os.path.isfile(starting_path):
        return os.path.getsize(starting_path)

    elif os.path.isdir(starting_path):
        size = 0
        for dirname, subdirs, files in os.walk(starting_path):
            if dirname.split("/")[-1] in EXCLUDE_SUBDIRS:
                continue

            for filename in files:
                full_path = os.path.join(dirname, filename)
                size += os.path.getsize(full_path)

        return size

    else:
        raise Exception


def compute_index(
    dependency_graph: weighted_dependency_graph,
    number_of_top_level_modules_to_remove: int = 2,
) -> Dict:
    index = {}
    _top_level_ids = [x.id for x in dependency_graph._top_level_nodes]
    _start_removals = []

    _start_removals = list(
        combinations(_top_level_ids, number_of_top_level_modules_to_remove)
    )

    for start_removal in _start_removals:
        t = set(start_removal)

        weight_removed = 0

        already_removed = set()
        for top_leaf_child in dependency_graph.HEAD.get_children():
            if top_leaf_child.id in t:
                t2 = t.union(set(["HEAD"]))
                weight_pruned, already_removed = _recursively_compute_removal(
                    top_leaf_child, t2, already_removed
                )
                weight_removed += weight_pruned
            else:
                weight_pruned, already_removed = _recursively_compute_removal(
                    top_leaf_child, t, already_removed
                )
                weight_removed += weight_pruned

        index[start_removal] = weight_removed

    return index


def _recursively_compute_removal(
    current_node: weighted_dependency_node,
    modules_removing: Set[str],
    already_removed: Set[str],
) -> Tuple[int, Set[str]]:
    if not current_node.get_parents():
        raise Exception

    if current_node.id in already_removed:
        # print(f"Already removed {current_node.id}")
        return 0, already_removed

    # All the parent ids and this node must be in the modules to remove if we are going to prune this node
    parent_ids = {x.id for x in current_node.get_parents()}

    remaining_parents = parent_ids.difference(modules_removing)
    # print(f"{current_node.id} has parents {parent_ids} and we are currently removing {modules_removing} -> {remaining_parents} ; {already_removed}")
    if len(remaining_parents) == 0:
        # IF all the parents of this node are in the removing set then remove this module weight
        weight = current_node.individual_weight
        already_removed.add(current_node.id)
        modules_removing.add(current_node.id)

        if current_node.get_children():
            for child in current_node.get_children():
                weight_pruned, already_removed = _recursively_compute_removal(
                    child, modules_removing, already_removed
                )
                weight += weight_pruned

        return weight, already_removed

    else:
        # IF there are other parents remaining then the removing set would not actually cause this module to be removed
        return 0, already_removed
